# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from blazar.notification import notifier

IMPL = notifier.Notifier()


def send_lease_notification(context, lease, notification):
    IMPL.send_lease_notification(context, lease, notification)


def format_lease_payload(lease):
    return {
        'lease_id': lease['id'],
        'user_id': lease['user_id'],
        'project_id': lease['project_id'],
        'start_date': lease['start_date'],
        'end_date': lease['end_date']
    }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from blazar.monitor import base

LOG = logging.getLogger(__name__)


class PollingMonitor(base.BaseMonitor):
    """A polling based monitor."""

    def __init__(self, monitor_plugins):
        """Initialize a polling monitor."""
        LOG.debug('Initializing a polling monitor...')
        super(PollingMonitor, self).__init__(monitor_plugins)
        self.polling_timers = []

    def start_monitoring(self):
        """Start polling."""
        LOG.debug('Starting a polling monitor...')

        try:
            for plugin in self.monitor_plugins:
                # Set polling timer. Wrap the monitor plugin method with the
                # call_monitor_plugin() to manage lease/reservation flags.
                self.polling_timers.append(
                    self.tg.add_timer(plugin.get_polling_interval(),
                                      self.call_monitor_plugin, None,
                                      plugin.poll))
            super(PollingMonitor, self).start_monitoring()
        except Exception as e:
            LOG.exception('Failed to start a polling monitor. (%s)',
                          str(e))

    def stop_monitoring(self):
        """Stop polling."""
        LOG.debug('Stopping a polling monitor...')
        try:
            for timer in self.polling_timers:
                self.tg.timer_done(timer)
            super(PollingMonitor, self).stop_monitoring()
        except Exception as e:
            LOG.exception('Failed to stop a polling monitor. (%s)', str(e))



# Copyright (c) 2017 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from blazar.manager import exceptions as manager_exceptions
from blazar import tests
from blazar.utils import plugins as plugins_utils


class TestPluginsUtils(tests.TestCase):

    def setUp(self):
        super(TestPluginsUtils, self).setUp()

    def test_convert_requirements_empty(self):
        request = '[]'
        result = plugins_utils.convert_requirements(request)
        self.assertEqual([], result)

    def test_convert_requirements_empty_string(self):
        # NOTE(priteau): Currently, empty requirements can be persisted in the
        # database as empty strings, which are not valid JSON objects.
        request = ''
        result = plugins_utils.convert_requirements(request)
        self.assertEqual([], result)

    def test_convert_requirements_small(self):
        request = '["=", "$memory", "4096"]'
        result = plugins_utils.convert_requirements(request)
        self.assertEqual(['memory == 4096'], result)

    def test_convert_requirements_with_incorrect_syntax_1(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements, '["a", "$memory", "4096"]')

    def test_convert_requirements_with_incorrect_syntax_2(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements, '["=", "memory", "4096"]')

    def test_convert_requirements_with_incorrect_syntax_3(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements, '["=", "$memory", 4096]')

    def test_convert_requirements_complex(self):
        request = '["and", [">", "$memory", "4096"], [">", "$disk", "40"]]'
        result = plugins_utils.convert_requirements(request)
        self.assertEqual(['memory > 4096', 'disk > 40'], result)

    def test_convert_requirements_complex_with_incorrect_syntax_1(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements,
            '["and", [">", "memory", "4096"], [">", "$disk", "40"]]')

    def test_convert_requirements_complex_with_incorrect_syntax_2(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements,
            '["fail", [">", "$memory", "4096"], [">", "$disk", "40"]]')

    def test_convert_requirements_complex_with_not_json_value(self):
        self.assertRaises(
            manager_exceptions.MalformedRequirements,
            plugins_utils.convert_requirements, 'something')

    def test_list_difference(self):
        old_list = [1, 1, 2, 3, 4, 4, 4, 5]
        new_list = [1, 2, 3, 4, 7, 8, 8]

        result = plugins_utils.list_difference(old_list, new_list)

        to_remove = [1, 4, 4, 5]
        to_add = [7, 8, 8]

        self.assertEqual((to_remove, to_add), result)

    def test_list_difference_empty(self):
        old_list = []
        new_list = [1, 2, 2, 2, 3, 4, 7, 8, 8]

        result = plugins_utils.list_difference(old_list, new_list)

        to_remove = []
        to_add = [1, 2, 2, 2, 3, 4, 7, 8, 8]

        self.assertEqual((to_remove, to_add), result)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_config import fixture as conf_fixture
from oslo_utils.fixture import uuidsentinel

from blazar import context
from blazar import tests

CONF = cfg.CONF


class TestBlazarContext(tests.TestCase):

    def setUp(self):
        super(TestBlazarContext, self).setUp()

        self.cfg = self.useFixture(conf_fixture.Config(CONF))
        self.cfg.config(os_admin_username='fake-admin')
        self.cfg.config(os_admin_user_domain_name='fake-admin-domain')
        self.cfg.config(os_admin_project_name='fake-admin-project')
        self.cfg.config(os_admin_project_domain_name='fake-admin-domain')

    def test_to_dict(self):
        ctx = context.BlazarContext(
            user_id=111, project_id=222,
            request_id='req-679033b7-1755-4929-bf85-eb3bfaef7e0b')

        # NOTE(priteau): for compatibility with oslo.context<4.0.0 which
        # returns a tenant key instead of project_id
        ctx_dict = ctx.to_dict()
        if 'tenant' in ctx_dict:
            ctx_dict['project_id'] = ctx_dict['tenant']
            del ctx_dict['tenant']

        expected = {
            'auth_token': None,
            'domain': None,
            'global_request_id': None,
            'is_admin': False,
            'is_admin_project': True,
            'project': 222,
            'project_domain': None,
            'project_id': 222,
            'read_only': False,
            'request_id': 'req-679033b7-1755-4929-bf85-eb3bfaef7e0b',
            'resource_uuid': None,
            'roles': [],
            'service_catalog': [],
            'show_deleted': False,
            'system_scope': None,
            'user': 111,
            'user_domain': None,
            'user_identity': '111 222 - - -'}
        self.assertEqual(expected, ctx_dict)

    def test_service_catalog_default(self):
        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id)
        self.assertEqual([], ctxt.service_catalog)

        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id,
                                     service_catalog=[])
        self.assertEqual([], ctxt.service_catalog)

        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id,
                                     service_catalog=None)
        self.assertEqual([], ctxt.service_catalog)

    def test_admin(self):
        ctx = context.admin()
        self.assertEqual(ctx.user_name, 'fake-admin')
        self.assertEqual(ctx.user_domain_name, 'fake-admin-domain')
        self.assertEqual(ctx.project_name, 'fake-admin-project')
        self.assertEqual(ctx.project_domain_name, 'fake-admin-domain')
        self.assertEqual(ctx.is_admin, True)

    def test_admin_nested(self):
        """Test that admin properties take priority over current context."""
        request_id = 'req-679033b7-1755-4929-bf85-eb3bfaef7e0b'
        service_catalog = ['foo']
        ctx = context.BlazarContext(
            user_name='fake-user', user_domain_name='fake-user-domain',
            project_name='fake-project',
            project_domain_name='fake-user-domain',
            service_catalog=service_catalog, request_id=request_id)
        with ctx:
            admin_ctx = context.admin()
            self.assertEqual(admin_ctx.user_name, 'fake-admin')
            self.assertEqual(admin_ctx.user_domain_name, 'fake-admin-domain')
            self.assertEqual(admin_ctx.project_name, 'fake-admin-project')
            self.assertEqual(admin_ctx.project_domain_name,
                             'fake-admin-domain')
            self.assertEqual(admin_ctx.is_admin, True)
            self.assertEqual(admin_ctx.request_id, request_id)
            self.assertEqual(admin_ctx.service_catalog, service_catalog)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.monitor import notification_monitor
from blazar.monitor import polling_monitor


def load_monitors(plugins):
    """Load resource monitors.

    :param plugins: resource plugins.
    :return: a list of monitors.
    """
    monitors = []

    # Setup a notification monitor
    notification_plugins = set([])
    for plugin in plugins.values():
        if plugin.monitor:
            if plugin.monitor.is_notification_enabled():
                notification_plugins.add(plugin.monitor)
    if notification_plugins:
        monitors.append(
            notification_monitor.NotificationMonitor(notification_plugins))

    # Setup a polling monitor
    polling_plugins = set([])
    for plugin in plugins.values():
        if plugin.monitor:
            if plugin.monitor.is_polling_enabled():
                polling_plugins.add(plugin.monitor)
    if polling_plugins:
        monitors.append(polling_monitor.PollingMonitor(polling_plugins))

    return monitors



# Copyright (c) 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt

from blazar.api.v1 import api_version_request
from blazar import exceptions
from blazar import tests


@ddt.ddt
class APIVersionRequestTests(tests.TestCase):
    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_version_request.APIVersionRequest(version)
            self.assertEqual(v._ver_major, exp_major)
            self.assertEqual(v._ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_min_version(self):
        self.assertEqual(
            api_version_request.APIVersionRequest(
                api_version_request.MIN_API_VERSION),
            api_version_request.min_api_version())

    def test_max_api_version(self):
        self.assertEqual(
            api_version_request.APIVersionRequest(
                api_version_request.MAX_API_VERSION),
            api_version_request.max_api_version())

    def test_null_version(self):
        v = api_version_request.APIVersionRequest()
        self.assertTrue(v.is_null())

    def test_not_null_version(self):
        v = api_version_request.APIVersionRequest('1.1')
        self.assertTrue(bool(v))

    @ddt.data("1", "100", "1.1.4", "100.23.66.3", "1 .1", "1. 1",
              "1.03", "01.1", "1.001", "", " 1.1", "1.1 ")
    def test_invalid_version_strings(self, version_string):
        self.assertRaises(exceptions.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest,
                          version_string)

    def test_version_comparisons(self):
        vers1 = api_version_request.APIVersionRequest("1.0")
        vers2 = api_version_request.APIVersionRequest("1.5")
        vers3 = api_version_request.APIVersionRequest("2.23")
        vers4 = api_version_request.APIVersionRequest("2.0")
        v_null = api_version_request.APIVersionRequest()

        self.assertLess(v_null, vers2)
        self.assertLess(vers1, vers2)
        self.assertLessEqual(vers1, vers2)
        self.assertLessEqual(vers1, vers4)
        self.assertGreater(vers2, v_null)
        self.assertGreater(vers3, vers2)
        self.assertGreaterEqual(vers4, vers1)
        self.assertGreaterEqual(vers3, vers2)
        self.assertNotEqual(vers1, vers2)
        self.assertNotEqual(vers1, vers4)
        self.assertNotEqual(vers1, v_null)
        self.assertEqual(v_null, v_null)
        self.assertRaises(TypeError, vers1.__lt__, "2.1")
        self.assertRaises(TypeError, vers1.__gt__, "2.1")
        self.assertRaises(TypeError, vers1.__eq__, "1.0")

    def test_version_matches(self):
        vers1 = api_version_request.APIVersionRequest("1.0")
        vers2 = api_version_request.APIVersionRequest("1.1")
        vers3 = api_version_request.APIVersionRequest("1.2")
        vers4 = api_version_request.APIVersionRequest("2.0")
        vers5 = api_version_request.APIVersionRequest("1.1")
        v_null = api_version_request.APIVersionRequest()

        self.assertTrue(vers2.matches(vers1, vers3))
        self.assertTrue(vers2.matches(v_null, vers5))
        self.assertTrue(vers2.matches(vers1, v_null))
        self.assertTrue(vers1.matches(v_null, v_null))
        self.assertFalse(vers2.matches(vers3, vers4))
        self.assertRaises(ValueError, v_null.matches, vers1, vers3)

    def test_get_string(self):
        vers1_string = "1.13"
        vers1 = api_version_request.APIVersionRequest(vers1_string)
        self.assertEqual(vers1_string, vers1.get_string())

        self.assertRaises(ValueError,
                          api_version_request.APIVersionRequest().get_string)

    @ddt.data(('1', '0'), ('1', '1'))
    @ddt.unpack
    def test_str(self, major, minor):
        request_input = '%s.%s' % (major, minor)
        request = api_version_request.APIVersionRequest(request_input)
        request_string = str(request)

        self.assertEqual('API Version Request '
                         'Major: %s, Minor: %s' % (major, minor),
                         request_string)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import retrying

from keystoneauth1 import adapter
from keystoneauth1.identity import v3
from keystoneauth1 import session

from oslo_config import cfg
from oslo_log import log as logging

from blazar import context
from blazar.utils.openstack import base
from blazar.utils.openstack import exceptions

placement_opts = [
    cfg.StrOpt('endpoint_type',
               default='internal',
               choices=['public', 'admin', 'internal'],
               help='Type of the placement endpoint to use. This endpoint '
                    'will be looked up in the keystone catalog and should be '
                    'one of public, internal or admin.'),
]

CONF = cfg.CONF
CONF.register_opts(placement_opts, group='placement')
LOG = logging.getLogger(__name__)

PLACEMENT_MICROVERSION = 1.29


class BlazarPlacementClient(object):
    """Client class for updating placement."""

    def _create_client(self, **kwargs):
        """Create the HTTP session accessing the placement service."""
        ctx = kwargs.pop('ctx', None)
        username = kwargs.pop('username',
                              CONF.os_admin_username)
        user_domain_name = kwargs.pop('user_domain_name',
                                      CONF.os_admin_user_domain_name)
        project_name = kwargs.pop('project_name',
                                  CONF.os_admin_project_name)
        password = kwargs.pop('password',
                              CONF.os_admin_password)

        project_domain_name = kwargs.pop('project_domain_name',
                                         CONF.os_admin_project_domain_name)
        auth_url = kwargs.pop('auth_url', None)
        region_name = kwargs.pop('region_name', CONF.os_region_name)

        if ctx is None:
            try:
                ctx = context.current()
            except RuntimeError:
                pass
        if ctx is not None:
            kwargs.setdefault('global_request_id', ctx.global_request_id)

        if auth_url is None:
            auth_url = "%s://%s:%s" % (CONF.os_auth_protocol,
                                       base.get_os_auth_host(CONF),
                                       CONF.os_auth_port)
            if CONF.os_auth_prefix:
                auth_url += "/%s" % CONF.os_auth_prefix
            if CONF.os_auth_version:
                auth_url += "/%s" % CONF.os_auth_version

        auth = v3.Password(auth_url=auth_url,
                           username=username,
                           password=password,
                           project_name=project_name,
                           user_domain_name=user_domain_name,
                           project_domain_name=project_domain_name)
        sess_kwargs = dict(
            auth=auth
        )
        if CONF.cafile:
            sess_kwargs.update(verify=CONF.cafile)
        sess = session.Session(**sess_kwargs)
        # Set accept header on every request to ensure we notify placement
        # service of our response body media type preferences.
        headers = {'accept': 'application/json'}
        kwargs.setdefault('service_type', 'placement')
        kwargs.setdefault('interface', CONF.placement.endpoint_type)
        kwargs.setdefault('additional_headers', headers)
        kwargs.setdefault('region_name', region_name)
        client = adapter.Adapter(sess, **kwargs)
        return client

    def get(self, url, microversion=PLACEMENT_MICROVERSION):
        client = self._create_client()
        return client.get(url, raise_exc=False,
                          microversion=microversion)

    def post(self, url, data, microversion=PLACEMENT_MICROVERSION):
        client = self._create_client()
        return client.post(url, json=data, raise_exc=False,
                           microversion=microversion)

    def put(self, url, data, microversion=PLACEMENT_MICROVERSION):
        client = self._create_client()
        return client.put(url, json=data, raise_exc=False,
                          microversion=microversion)

    def delete(self, url, microversion=PLACEMENT_MICROVERSION):
        client = self._create_client()
        return client.delete(url, raise_exc=False,
                             microversion=microversion)

    def get_resource_provider(self, rp_name):
        """Calls the placement API for a resource provider record.

        :param rp_name: Name of the resource provider
        :return: A dict of resource provider information
                 or None if the resource provider doesn't exist.
        :raise: ResourceProviderRetrievalFailed on error.
        """
        url = "/resource_providers?name=%s" % rp_name
        resp = self.get(url)
        if resp:
            json_resp = resp.json()
            if json_resp['resource_providers']:
                return json_resp['resource_providers'][0]
            else:
                return None

        msg = ("Failed to get resource provider %(name)s. "
               "Got %(status_code)d: %(err_text)s.")
        args = {
            'name': rp_name,
            'status_code': resp.status_code,
            'err_text': resp.text,
        }
        LOG.error(msg, args)
        raise exceptions.ResourceProviderRetrievalFailed(name=rp_name)

    def create_resource_provider(self, rp_name, rp_uuid=None,
                                 parent_uuid=None):
        """Calls the placement API to create a new resource provider record.

        :param rp_name: Name of the resource provider
        :param rp_uuid: Optional UUID of the new resource provider
        :param parent_uuid: Optional UUID of the parent resource provider
        :return: A dict of resource provider information object representing
                 the newly-created resource provider.
        :raise: ResourceProviderCreationFailed error.
        """
        url = "/resource_providers"
        payload = {'name': rp_name}
        if rp_uuid is not None:
            payload['uuid'] = rp_uuid
        if parent_uuid is not None:
            payload['parent_provider_uuid'] = parent_uuid

        resp = self.post(url, payload)

        if resp:
            msg = ("Created resource provider record via placement API for "
                   "resource provider %(name)s.")
            args = {'name': rp_name}
            LOG.info(msg, args)
            return resp.json()

        if resp.status_code == 409:
            msg = ("Conflict on creating resource provider %(name)s in "
                   "placement API. Got %(status_code)d: %(err_text)s.")
            args = {
                'name': rp_name,
                'status_code': resp.status_code,
                'err_text': resp.text,
            }
            LOG.error(msg, args)
            raise exceptions.ResourceProviderCreationConflict(name=rp_name)

        msg = ("Failed to create resource provider record in placement API "
               "for resource provider %(name)s. "
               "Got %(status_code)d: %(err_text)s.")
        args = {
            'name': rp_name,
            'status_code': resp.status_code,
            'err_text': resp.text,
        }
        LOG.error(msg, args)
        raise exceptions.ResourceProviderCreationFailed(name=rp_name)

    def delete_resource_provider(self, rp_uuid):
        """Calls the placement API to delete a resource provider.

        :param rp_uuid: UUID of the resource provider to delete
        :raise: ResourceProviderDeletionFailed error
        """
        url = '/resource_providers/%s' % rp_uuid
        resp = self.delete(url)

        if resp:
            LOG.info("Deleted resource provider %s", rp_uuid)
            return

        msg = ("Failed to delete resource provider with UUID %(uuid)s from "
               "the placement API. Got %(status_code)d: %(err_text)s.")
        args = {
            'uuid': rp_uuid,
            'status_code': resp.status_code,
            'err_text': resp.text
        }
        LOG.error(msg, args)
        raise exceptions.ResourceProviderDeletionFailed(uuid=rp_uuid)

    def create_reservation_provider(self, host_name):
        """Create a reservation provider as a child of the given host"""
        host_rp = self.get_resource_provider(host_name)
        if host_rp is None:
            raise exceptions.ResourceProviderNotFound(
                resource_provider=host_name)
        host_uuid = host_rp['uuid']
        rp_name = "blazar_" + host_name

        reservation_rp = self.create_resource_provider(
            rp_name, parent_uuid=host_uuid)
        return reservation_rp

    def delete_reservation_provider(self, host_name):
        """Delete the reservation provider, the child of the given host"""
        rp_name = "blazar_" + host_name
        rp = self.get_resource_provider(rp_name)
        if rp is None:
            # If the reservation provider doesn't exist,
            # no operation will be performed.
            return
        rp_uuid = rp['uuid']
        self.delete_resource_provider(rp_uuid)

    def create_resource_class(self, rc_name):
        """Calls the placement API to create a resource class.

        :param rc_name: string name of the resource class to create. This
                        shall be something like "CUSTOM_RESERVATION_{uuid}".
        :raises: ResourceClassCreationFailed error.
        """
        url = '/resource_classes'
        payload = {'name': rc_name}
        resp = self.post(url, payload)
        if resp:
            LOG.info("Created resource class %s", rc_name)
            return
        msg = ("Failed to create resource class with placement API for "
               "%(rc_name)s. Got %(status_code)d: %(err_text)s.")
        args = {
            'rc_name': rc_name,
            'status_code': resp.status_code,
            'err_text': resp.text,
        }
        LOG.error(msg, args)
        raise exceptions.ResourceClassCreationFailed(resource_class=rc_name)

    def delete_resource_class(self, rc_name):
        """Calls the placement API to delete a resource class.

        :param rc_name: string name of the resource class to delete. This
                        shall be something like "CUSTOM_RESERVATION_{uuid}"
        :raises: ResourceClassDeletionFailed error.
        """
        url = '/resource_classes/%s' % rc_name
        resp = self.delete(url)
        if resp:
            LOG.info("Deleted resource class %s", rc_name)
            return
        msg = ("Failed to delete resource class with placement API for "
               "%(rc_name)s. Got %(status_code)d: %(err_text)s.")
        args = {
            'rc_name': rc_name,
            'status_code': resp.status_code,
            'err_text': resp.text,
        }
        LOG.error(msg, args)
        raise exceptions.ResourceClassDeletionFailed(resource_class=rc_name)

    def create_reservation_class(self, reservation_uuid):
        """Create the reservation class from the given reservation uuid"""
        # Placement API doesn't accept resource classes with lower characters
        # and "-"(hyphen) in its name. We should translate the uuid here.
        reservation_uuid = reservation_uuid.upper().replace("-", "_")
        rc_name = 'CUSTOM_RESERVATION_' + reservation_uuid
        self.create_resource_class(rc_name)

    def delete_reservation_class(self, reservation_uuid):
        """Delete the reservation class from the given reservation uuid"""
        # Placement API doesn't accept resource classes with lower characters
        # and "-"(hyphen) in its name. We should translate the uuid here.
        reservation_uuid = reservation_uuid.upper().replace("-", "_")
        rc_name = 'CUSTOM_RESERVATION_' + reservation_uuid
        try:
            self.delete_resource_class(rc_name)
        except exceptions.ResourceClassDeletionFailed:
            # We just log it and skip to keep the compatibility before Stein
            LOG.info("Resource class %s doesn't exist. Skipped the deletion "
                     "of the resource class", rc_name)

    def get_inventory(self, rp_uuid):
        """Calls the placement API to get resource inventory information.

        :param rp_uuid: UUID of the resource provider to get
        """
        url = '/resource_providers/%s/inventories' % rp_uuid
        resp = self.get(url)
        if resp:
            return resp.json()
        raise exceptions.ResourceProviderNotFound(resource_provider=rp_uuid)

    def get_traits(self, rp_uuid):
        """Calls the placement API to get resource trait information.

        :param rp_uuid: UUID of the resource provider to get
        """
        url = '/resource_providers/%s/traits' % rp_uuid
        resp = self.get(url)
        if resp:
            return resp.json().get("traits", [])
        raise exceptions.ResourceProviderNotFound(resource_provider=rp_uuid)

    @retrying.retry(stop_max_attempt_number=5,
                    retry_on_exception=lambda e: isinstance(
                        e, exceptions.InventoryConflict))
    def update_inventory(self, rp_uuid, rc_name, num, additional):
        """Update the inventory for the resource provider.

        :param rp_uuid: The resource provider UUID for the operation
        :param rc_name: The resource class name of the inventory to update
        :param num: The total inventory to add/update
        :param additional: Add the given number amounts to the existing if
                           True, else just overwrite the total value
        :raises: ResourceProviderNotFound or InventoryUpdateFailed error.
        """
        curr = self.get_inventory(rp_uuid)
        inventories = curr['inventories']
        generation = curr['resource_provider_generation']

        if additional and rc_name in inventories:
            inventories[rc_name]["total"] += num

        else:
            inv_data = {
                rc_name: {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": num
                },
            }
            inventories.update(inv_data)

        payload = {
            'inventories': inventories,
            'resource_provider_generation': generation,
        }
        url = '/resource_providers/%s/inventories' % rp_uuid

        resp = self.put(url, payload)
        if resp:
            return resp.json()

        if resp.status_code == 409:
            err = resp.json()['errors'][0]
            if err['code'] == 'placement.concurrent_update':
                # NOTE(tetsuro): Another thread updated the inventory of the
                # same rp during the get_inventory() and the put(). We simply
                # retry it for this case.
                msg = ("Conflict on updating inventory in placement. "
                       "Got %(status_code)d: %(err_text)s. ")
                args = {
                    'status_code': resp.status_code,
                    'err_text': resp.text,
                }
                LOG.error(msg, args)
                raise exceptions.InventoryConflict(resource_provider=rp_uuid)

        raise exceptions.InventoryUpdateFailed(resource_provider=rp_uuid)

    def delete_inventory(self, rp_uuid, rc_name):
        """Delete the inventory for the resource provider.

        :param rp_uuid: The resource provider UUID for the operation
        :param rc_name: The resource class name to delete from inventory
        :raises: InventoryUpdateFailed error
        """
        url = '/resource_providers/%s/inventories/%s' % (rp_uuid, rc_name)

        resp = self.delete(url)
        if resp:
            return

        raise exceptions.InventoryUpdateFailed(resource_provider=rp_uuid)

    def update_reservation_inventory(self, host_name, reserv_uuid, num,
                                     additional=False):
        """Update the reservation inventory for the reservation provider.

        :param host_name: The name of the target host
        :param reserv_uuid: The reservation uuid
        :param num: The number of the instances to reserve on the host
        :return: The updated inventory record
        """
        # Get reservation provider uuid
        rp_name = "blazar_" + host_name
        rp = self.get_resource_provider(rp_name)
        if rp is None:
            # If the reservation provider is not created yet,
            # this function creates it.
            rp = self.create_reservation_provider(host_name)
        rp_uuid = rp['uuid']

        # Get resource class name
        reserv_uuid = reserv_uuid.upper().replace("-", "_")
        rc_name = 'CUSTOM_RESERVATION_' + reserv_uuid

        return self.update_inventory(rp_uuid, rc_name, num, additional)

    def delete_reservation_inventory(self, host_name, reserv_uuid):
        """Delete the reservation inventory for the reservation provider.

        :param host_name: The name of the target host
        :param reserv_uuid: The reservation uuid
        :raises: ResourceProviderNotFound if the reservation
                 provider is not found
        """
        # Get reservation provider uuid
        rp_name = "blazar_" + host_name
        rp = self.get_resource_provider(rp_name)
        if rp is None:
            raise exceptions.ResourceProviderNotFound(
                resource_provider=rp_name)
        rp_uuid = rp['uuid']

        # Convert reservation uuid to resource class name
        reserv_uuid = reserv_uuid.upper().replace("-", "_")
        rc_name = 'CUSTOM_RESERVATION_' + reserv_uuid
        try:
            self.delete_inventory(rp_uuid, rc_name)
        except exceptions.InventoryUpdateFailed:
            # We just log it and skip to keep the compatibility before Stein
            LOG.info("Resource class %s doesn't exist or there is no "
                     "inventory for that resource class on resource provider "
                     "%s. Skipped the deletion", rc_name, rp_name)



# Copyright (c) 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from blazar import exceptions
from blazar.i18n import _


LOG = logging.getLogger(__name__)


class BlazarDBException(exceptions.BlazarException):
    msg_fmt = _('An unknown database exception occurred')


class BlazarDBDuplicateEntry(BlazarDBException):
    msg_fmt = _('Duplicate entry for %(columns)s in %(model)s model was found')


class BlazarDBNotFound(BlazarDBException):
    msg_fmt = _('%(id)s %(model)s was not found')


class BlazarDBInvalidFilter(BlazarDBException):
    msg_fmt = _('%(query_filter)s is invalid')


class BlazarDBInvalidFilterOperator(BlazarDBException):
    msg_fmt = _('%(filter_operator)s is invalid')


class BlazarDBResourcePropertiesNotEnabled(BlazarDBException):
    msq_fmt = _('%(resource_type)s does not have resource properties enabled.')


class BlazarDBInvalidResourceProperty(BlazarDBException):
    msg_fmt = _('%(property_name)s does not exist for resource type '
                '%(resource_type)s.')



# -*- coding: utf-8 -*-
#
# Author: François Rossigneux <francois.rossigneux@inria.fr>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_context import context
from oslo_utils import uuidutils

from blazar.db.sqlalchemy import api as db_api
from blazar.db.sqlalchemy import utils as db_utils
from blazar.manager import exceptions as mgr_exceptions
from blazar import tests


def _get_fake_random_uuid():
    return str(uuidutils.generate_uuid())


def _get_fake_lease_uuid():
    """Returns a fake uuid."""
    return 'aaaaaaaa-1111-bbbb-2222-cccccccccccc'


def _get_fake_phys_reservation_values(lease_id=_get_fake_lease_uuid(),
                                      resource_id='1234'):
    return {'lease_id': lease_id,
            'resource_id': resource_id,
            'resource_type': 'physical:host'}


def _get_fake_inst_reservation_values(lease_id=_get_fake_lease_uuid(),
                                      resource_id='5678'):
    return {'lease_id': lease_id,
            'resource_id': resource_id,
            'resource_type': 'virtual:instance'}


def _get_datetime(value='2030-01-01 00:00'):
    return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M')


def _get_fake_phys_lease_values(id=_get_fake_lease_uuid(),
                                name='fake_phys_lease',
                                start_date=_get_datetime('2030-01-01 00:00'),
                                end_date=_get_datetime('2030-01-02 00:00'),
                                resource_id=None):
    return {'id': id,
            'name': name,
            'start_date': start_date,
            'end_date': end_date,
            'trust': 'trust',
            'reservations': [_get_fake_phys_reservation_values(
                lease_id=id,
                resource_id=resource_id)],
            'events': []
            }


def _create_physical_lease(values=_get_fake_phys_lease_values(),
                           random=False):
    """Creating fake lease having a single physical resource."""
    if random is True:
        values = _get_fake_phys_lease_values(id=_get_fake_random_uuid(),
                                             name=_get_fake_random_uuid())
    lease = db_api.lease_create(values)
    for reservation in db_api.reservation_get_all_by_lease_id(lease['id']):
        allocation_values = {
            'id': _get_fake_random_uuid(),
            'compute_host_id': reservation['resource_id'],
            'reservation_id': reservation['id']
        }
        db_api.host_allocation_create(allocation_values)
    return lease


class SQLAlchemyDBUtilsTestCase(tests.DBTestCase):
    """Test case for SQLAlchemy DB utils."""

    def setUp(self):
        super(SQLAlchemyDBUtilsTestCase, self).setUp()
        self.set_context(context.get_admin_context())

    def _setup_leases(self):
        """Setup some leases."""
        r1 = _get_fake_phys_lease_values(
            id='lease1',
            name='fake_phys_lease_r1',
            start_date=_get_datetime('2030-01-01 09:00'),
            end_date=_get_datetime('2030-01-01 10:30'),
            resource_id='r1')
        r2 = _get_fake_phys_lease_values(
            id='lease2',
            name='fake_phys_lease_r2',
            start_date=_get_datetime('2030-01-01 11:00'),
            end_date=_get_datetime('2030-01-01 12:45'),
            resource_id='r2')
        r3 = _get_fake_phys_lease_values(
            id='lease3',
            name='fake_phys_lease_r3',
            start_date=_get_datetime('2030-01-01 13:00'),
            end_date=_get_datetime('2030-01-01 14:00'),
            resource_id='r1')
        _create_physical_lease(values=r1)
        _create_physical_lease(values=r2)
        _create_physical_lease(values=r3)

    def check_reservation(self, expect, host_ids, start, end):
        expect.sort(key=lambda x: x['lease_id'])
        if isinstance(host_ids, list):
            ret = db_utils.get_reservations_by_host_ids(host_ids, start, end)
        else:
            ret = db_utils.get_reservations_by_host_id(host_ids, start, end)

        for i, res in enumerate(sorted(ret, key=lambda x: x['lease_id'])):
            self.assertEqual(expect[i]['lease_id'], res['lease_id'])
            self.assertEqual(expect[i]['resource_id'], res['resource_id'])
            self.assertEqual(expect[i]['resource_type'], res['resource_type'])

    def test_get_free_periods(self):
        """Find the free periods."""
        self._setup_leases()
        start_date = datetime.datetime.strptime('2028-01-01 08:00',
                                                '%Y-%m-%d %H:%M')
        end_date = datetime.datetime.strptime('2099-01-01 00:00',
                                              '%Y-%m-%d %H:%M')
        duration = datetime.timedelta(hours=1)
        free_periods = db_utils.get_free_periods('r1',
                                                 start_date,
                                                 end_date,
                                                 duration)
        self.assertEqual(3, len(free_periods))
        self.assertEqual('2028-01-01 08:00',
                         free_periods[0][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 09:00',
                         free_periods[0][1].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 10:30',
                         free_periods[1][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 13:00',
                         free_periods[1][1].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 14:00',
                         free_periods[2][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2099-01-01 00:00',
                         free_periods[2][1].strftime('%Y-%m-%d %H:%M'))
        duration = datetime.timedelta(hours=3)
        free_periods = db_utils.get_free_periods('r1',
                                                 start_date,
                                                 end_date,
                                                 duration)
        self.assertEqual(2, len(free_periods))
        self.assertEqual('2028-01-01 08:00',
                         free_periods[0][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 09:00',
                         free_periods[0][1].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 14:00',
                         free_periods[1][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2099-01-01 00:00',
                         free_periods[1][1].strftime('%Y-%m-%d %H:%M'))

    def test_get_reserved_periods(self):
        """Find the reserved periods."""
        self._setup_leases()
        start_date = datetime.datetime.strptime('2028-01-01 08:00',
                                                '%Y-%m-%d %H:%M')
        end_date = datetime.datetime.strptime('2099-01-01 00:00',
                                              '%Y-%m-%d %H:%M')
        duration = datetime.timedelta(hours=1)
        reserved_periods = db_utils.get_reserved_periods('r1',
                                                         start_date,
                                                         end_date,
                                                         duration)
        self.assertEqual(2, len(reserved_periods))
        self.assertEqual('2030-01-01 09:00',
                         reserved_periods[0][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 10:30',
                         reserved_periods[0][1].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 13:00',
                         reserved_periods[1][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 14:00',
                         reserved_periods[1][1].strftime('%Y-%m-%d %H:%M'))
        duration = datetime.timedelta(hours=3)
        reserved_periods = db_utils.get_reserved_periods('r1',
                                                         start_date,
                                                         end_date,
                                                         duration)
        self.assertEqual(1, len(reserved_periods))
        self.assertEqual('2030-01-01 09:00',
                         reserved_periods[0][0].strftime('%Y-%m-%d %H:%M'))
        self.assertEqual('2030-01-01 14:00',
                         reserved_periods[0][1].strftime('%Y-%m-%d %H:%M'))

    def test_get_reservations_by_host_id(self):
        self._setup_leases()

        self.check_reservation([], 'r1',
                               '2030-01-01 07:00', '2030-01-01 08:59')

        ret = db_api.reservation_get_all_by_lease_id('lease1')
        self.check_reservation(ret, 'r1',
                               '2030-01-01 08:00', '2030-01-01 10:00')

        ret = db_api.reservation_get_all_by_lease_id('lease1')
        ret.extend(db_api.reservation_get_all_by_lease_id('lease3'))
        self.check_reservation(ret, 'r1',
                               '2030-01-01 08:00', '2030-01-01 15:30')

        self.check_reservation([], 'r4',
                               '2030-01-01 07:00', '2030-01-01 15:00')

    def test_get_reservations_by_host_id_with_multi_reservation(self):
        self._setup_leases()

        fake_lease = _get_fake_phys_lease_values(
            id='lease-4',
            name='fake_phys_lease_r4',
            start_date=_get_datetime('2030-01-01 15:00'),
            end_date=_get_datetime('2030-01-01 16:00'),
            resource_id='r4-1')

        fake_lease['reservations'].append(
            _get_fake_phys_reservation_values(lease_id='lease-4',
                                              resource_id='r1'))
        _create_physical_lease(values=fake_lease)

        expected = db_api.reservation_get_all_by_values(
            **{'resource_id': 'r1'})
        self.assertEqual(3, len(expected))
        self.check_reservation(expected, 'r1',
                               '2030-01-01 08:00', '2030-01-01 17:00')

    def test_get_reservations_by_host_ids(self):
        self._setup_leases()

        self.check_reservation([], ['r1', 'r2'],
                               '2030-01-01 07:00', '2030-01-01 08:59')

        ret = db_api.reservation_get_all_by_lease_id('lease1')
        self.check_reservation(ret, ['r1', 'r2'],
                               '2030-01-01 08:00', '2030-01-01 10:00')

        ret = db_api.reservation_get_all_by_lease_id('lease1')
        ret.extend(db_api.reservation_get_all_by_lease_id('lease2'))
        ret.extend(db_api.reservation_get_all_by_lease_id('lease3'))
        self.check_reservation(ret, ['r1', 'r2'],
                               '2030-01-01 08:00', '2030-01-01 15:30')

        self.check_reservation([], ['r4'],
                               '2030-01-01 07:00', '2030-01-01 15:00')

    def test_get_reservation_allocations_by_host_ids(self):
        self._setup_leases()

        # query all allocations of lease1, lease2 and lease3
        expected = [
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease1")[0]["id"],
                "status": None,
                "lease_id": "lease1",
                "start_date": datetime.datetime(2030, 1, 1, 9, 0),
                "end_date": datetime.datetime(2030, 1, 1, 10, 30),
                "lease_name": "fake_phys_lease_r1",
                "project_id": None,
                "host_ids": ["r1"]
            },
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease2")[0]["id"],
                "status": None,
                "lease_id": "lease2",
                "start_date": datetime.datetime(2030, 1, 1, 11, 0),
                "end_date": datetime.datetime(2030, 1, 1, 12, 45),
                "lease_name": "fake_phys_lease_r2",
                "project_id": None, "host_ids": ["r2"]
            },
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease3")[0]["id"],
                "status": None,
                "lease_id": "lease3",
                "start_date": datetime.datetime(2030, 1, 1, 13, 0),
                "end_date": datetime.datetime(2030, 1, 1, 14, 0),
                "lease_name": "fake_phys_lease_r3",
                "project_id": None, "host_ids": ["r1"]
            }
        ]
        ret = db_utils.get_reservation_allocations_by_host_ids(
            ['r1', 'r2'], '2030-01-01 08:00', '2030-01-01 15:00')
        self.assertListEqual(expected, ret)

        # query allocations of lease2 and lease3
        expected = [
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease2")[0]["id"],
                "status": None,
                "lease_id": "lease2",
                "start_date": datetime.datetime(2030, 1, 1, 11, 0),
                "end_date": datetime.datetime(2030, 1, 1, 12, 45),
                "lease_name": "fake_phys_lease_r2",
                "project_id": None, "host_ids": ["r2"]
            },
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease3")[0]["id"],
                "status": None,
                "lease_id": "lease3",
                "start_date": datetime.datetime(2030, 1, 1, 13, 0),
                "end_date": datetime.datetime(2030, 1, 1, 14, 0),
                "lease_name": "fake_phys_lease_r3",
                "project_id": None, "host_ids": ["r1"]
            }
        ]
        ret = db_utils.get_reservation_allocations_by_host_ids(
            ['r1', 'r2'], '2030-01-01 11:30', '2030-01-01 15:00')

        self.assertListEqual(expected, ret)

    def test_get_reservation_allocations_by_host_ids_with_lease_id(self):
        self._setup_leases()

        # query all allocations of lease1, lease2 and lease3
        expected = [
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease1")[0]["id"],
                "status": None,
                "lease_id": "lease1",
                "start_date": datetime.datetime(2030, 1, 1, 9, 0),
                "end_date": datetime.datetime(2030, 1, 1, 10, 30),
                "lease_name": "fake_phys_lease_r1",
                "project_id": None,
                "host_ids": ["r1"]
            },
        ]
        ret = db_utils.get_reservation_allocations_by_host_ids(
            ['r1', 'r2'], '2030-01-01 08:00', '2030-01-01 15:00', 'lease1')

        self.assertListEqual(expected, ret)

    def test_get_reservation_allocations_by_host_ids_with_reservation_id(self):
        self._setup_leases()
        reservation1 = db_api.reservation_get_all_by_lease_id('lease1')[0]

        # query allocations of lease1
        expected = [
            {
                "id": db_api.reservation_get_all_by_lease_id(
                    "lease1")[0]["id"],
                "status": None,
                "lease_id": "lease1",
                "start_date": datetime.datetime(2030, 1, 1, 9, 0),
                "end_date": datetime.datetime(2030, 1, 1, 10, 30),
                "lease_name": "fake_phys_lease_r1",
                "project_id": None,
                "host_ids": ["r1"]
            },
        ]
        ret = db_utils.get_reservation_allocations_by_host_ids(
            ['r1', 'r2'], '2030-01-01 08:00', '2030-01-01 15:00',
            reservation_id=reservation1['id'])

        self.assertListEqual(expected, ret)

    def test_get_plugin_reservation_with_host(self):
        patch_host_reservation_get = self.patch(db_api, 'host_reservation_get')
        patch_host_reservation_get.return_value = {
            'id': 'id',
            'reservation_id': 'reservation-id',
        }
        db_utils.get_plugin_reservation('physical:host', 'id-1')
        patch_host_reservation_get.assert_called_once_with('id-1')

    def test_get_plugin_reservation_with_instance(self):
        patch_inst_reservation_get = self.patch(db_api,
                                                'instance_reservation_get')
        patch_inst_reservation_get.return_value = {
            'id': 'id',
            'reservation_id': 'reservation-id',
        }
        db_utils.get_plugin_reservation('virtual:instance', 'id-1')
        patch_inst_reservation_get.assert_called_once_with('id-1')

    def test_get_plugin_reservation_with_invalid(self):
        self.assertRaises(mgr_exceptions.UnsupportedResourceType,
                          db_utils.get_plugin_reservation, 'invalid', 'id1')



# -*- coding: utf-8 -*-
#
# Author: François Rossigneux <francois.rossigneux@inria.fr>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict
import sys

import sqlalchemy as sa

from blazar.db.sqlalchemy import api
from blazar.db.sqlalchemy import facade_wrapper
from blazar.db.sqlalchemy import models
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins import instances as instance_plugin
from blazar.plugins import oshosts as host_plugin


def get_backend():
    """The backend is this module itself."""
    return sys.modules[__name__]


def _get_leases_from_resource_id(resource_id, start_date, end_date):
    with facade_wrapper.session_for_read() as session:
        border0 = start_date <= models.Lease.end_date
        border1 = models.Lease.start_date <= end_date
        query = (session.query(models.Lease).join(models.Reservation)
                 .filter(models.Reservation.resource_id == resource_id)
                 .filter(sa.and_(border0, border1)))
        for lease in query:
            yield lease


def _get_leases_from_host_id(host_id, start_date, end_date):
    with facade_wrapper.session_for_read() as session:
        border0 = start_date <= models.Lease.end_date
        border1 = models.Lease.start_date <= end_date
        query = (session.query(models.Lease).join(models.Reservation)
                 .join(models.ComputeHostAllocation)
                 .filter(models.ComputeHostAllocation.compute_host_id ==
                         host_id)
                 .filter(sa.and_(border0, border1)))
        for lease in query:
            yield lease


def _get_leases_from_fip_id(fip_id, start_date, end_date):
    with facade_wrapper.session_for_read() as session:
        border0 = sa.and_(models.Lease.start_date < start_date,
                          models.Lease.end_date < start_date)
        border1 = sa.and_(models.Lease.start_date > end_date,
                          models.Lease.end_date > end_date)
        query = (session.query(models.Lease).join(models.Reservation)
                 .join(models.FloatingIPAllocation)
                 .filter(models.FloatingIPAllocation.floatingip_id == fip_id)
                 .filter(~sa.or_(border0, border1)))
        for lease in query:
            yield lease


def get_reservations_by_host_id(host_id, start_date, end_date):
    with facade_wrapper.session_for_read() as session:
        border0 = start_date <= models.Lease.end_date
        border1 = models.Lease.start_date <= end_date
        query = (session.query(models.Reservation).join(models.Lease)
                 .join(models.ComputeHostAllocation)
                 .filter(models.ComputeHostAllocation.compute_host_id ==
                         host_id)
                 .filter(sa.and_(border0, border1)))
        return query.all()


def get_reservations_by_host_ids(host_ids, start_date, end_date):
    with facade_wrapper.session_for_read() as session:
        border0 = start_date <= models.Lease.end_date
        border1 = models.Lease.start_date <= end_date
        query = (session.query(models.Reservation).join(models.Lease)
                 .join(models.ComputeHostAllocation)
                 .filter(models.ComputeHostAllocation.compute_host_id
                         .in_(host_ids))
                 .filter(sa.and_(border0, border1)))
        return query.all()


def get_reservations_for_allocations(session, start_date, end_date,
                                     lease_id=None, reservation_id=None):
    fields = ['id', 'status', 'lease_id', 'start_date',
              'end_date', 'lease_name', 'project_id']

    reservations_query = (session.query(
        models.Reservation.id,
        models.Reservation.status,
        models.Reservation.lease_id,
        models.Lease.start_date,
        models.Lease.end_date,
        models.Lease.name,
        models.Lease.project_id)
        .join(models.Lease))

    if lease_id:
        reservations_query = reservations_query.filter(
            models.Reservation.lease_id == lease_id)
    if reservation_id:
        reservations_query = reservations_query.filter(
            models.Reservation.id == reservation_id)

    # Only enforce time restrictions if we're not targeting a specific
    # lease or reservation.
    if not (lease_id or reservation_id):
        border0 = models.Lease.end_date >= start_date
        border1 = models.Lease.start_date <= end_date
        reservations_query = reservations_query.filter(
            sa.and_(border0, border1))

    return [dict(zip(fields, r)) for r in reservations_query.all()]


def get_reservation_allocations_by_host_ids(host_ids, start_date, end_date,
                                            lease_id=None,
                                            reservation_id=None):
    with facade_wrapper.session_for_read() as session:
        # Get all reservations applicable
        reservations = get_reservations_for_allocations(
            session, start_date, end_date, lease_id, reservation_id)

        # Select (reservation_id, host_id) for all reservations
        allocations_query = (session.query(
            models.ComputeHostAllocation.reservation_id,
            models.ComputeHostAllocation.compute_host_id)
            .filter(models.ComputeHostAllocation.compute_host_id.in_(host_ids))
            .filter(models.ComputeHostAllocation.reservation_id.in_(
                list(set([x['id'] for x in reservations])))))

        # Create a mapping of reservation_id to list of host_ids
        allocations = defaultdict(list)
        for row in allocations_query.all():
            allocations[row[0]].append(row[1])

        # Copy the host id lists to the corresponding reservation
        for r in reservations:
            r['host_ids'] = allocations[r['id']]
        return reservations


def get_reservation_allocations_by_fip_ids(fip_ids, start_date, end_date,
                                           lease_id=None, reservation_id=None):
    with facade_wrapper.session_for_read() as session:
        reservations = get_reservations_for_allocations(
            session, start_date, end_date, lease_id, reservation_id)

        allocations_query = (session.query(
            models.FloatingIPAllocation.reservation_id,
            models.FloatingIPAllocation.floatingip_id)
            .filter(models.FloatingIPAllocation.floatingip_id.in_(fip_ids))
            .filter(models.FloatingIPAllocation.reservation_id.in_(
                list(set([x['id'] for x in reservations])))))

        allocations = defaultdict(list)

        for row in allocations_query.all():
            allocations[row[0]].append(row[1])

        for r in reservations:
            r['floatingip_ids'] = allocations[r['id']]

        return reservations


def get_plugin_reservation(resource_type, resource_id):
    if resource_type == host_plugin.RESOURCE_TYPE:
        return api.host_reservation_get(resource_id)
    elif resource_type == instance_plugin.RESOURCE_TYPE:
        return api.instance_reservation_get(resource_id)
    else:
        raise mgr_exceptions.UnsupportedResourceType(
            resource_type=resource_type)


def get_free_periods(resource_id, start_date, end_date, duration,
                     resource_type='host'):
    """Returns a list of free periods."""
    reserved_periods = get_reserved_periods(resource_id,
                                            start_date,
                                            end_date,
                                            duration,
                                            resource_type=resource_type)
    free_periods = []
    previous = (start_date, start_date)
    if len(reserved_periods) >= 1:
        for period in reserved_periods:
            free_periods.append((previous[1], period[0]))
            previous = period
        free_periods.append((previous[1], end_date))
        if free_periods[0][0] == free_periods[0][1]:
            del free_periods[0]
        if free_periods[-1][0] == free_periods[-1][1]:
            del free_periods[-1]
    elif start_date != end_date and start_date + duration <= end_date:
        free_periods.append((start_date, end_date))
    return free_periods


def _get_events(resource_id, start_date, end_date, resource_type):
    """Create a list of events."""
    events = {}
    if resource_type == 'host':
        leases = _get_leases_from_host_id(resource_id, start_date, end_date)
    elif resource_type == 'floatingip':
        leases = _get_leases_from_fip_id(resource_id, start_date, end_date)
    else:
        mgr_exceptions.UnsupportedResourceType(resource_type)

    for lease in leases:
        if lease.start_date < start_date:
            min_date = start_date
        else:
            min_date = lease.start_date
        if lease.end_date > end_date:
            max_date = end_date
        else:
            max_date = lease.end_date
        if min_date in events.keys():
            events[min_date]['quantity'] += 1
        else:
            events[min_date] = {'quantity': 1}
        if max_date in events.keys():
            events[max_date]['quantity'] -= 1
        else:
            events[max_date] = {'quantity': -1}
    return events


def _find_reserved_periods(events, quantity, capacity):
    """Find the reserved periods."""
    reserved_periods = []
    used = 0
    reserved_start = None
    for event_date in sorted(events):
        used += events[event_date]['quantity']
        if not reserved_start and used + quantity > capacity:
            reserved_start = event_date
        elif reserved_start and used + quantity <= capacity:
            reserved_periods.append((reserved_start, event_date))
            reserved_start = None
    return reserved_periods


def _merge_periods(reserved_periods, start_date, end_date, duration):
    """Merge periods if the interval is too narrow."""
    reserved_start = None
    reserved_end = None
    previous = []
    merged_reserved_periods = []
    for period in reserved_periods:
        if not reserved_start:
            reserved_start = period[0]
        # Enough time between the two reserved periods
        if previous and period[0] - previous[1] >= duration:
            reserved_end = previous[1]
            merged_reserved_periods.append((reserved_start, reserved_end))
            reserved_start = period[0]
        reserved_end = period[1]
        previous = period
    if previous and end_date - previous[1] < duration:
        merged_reserved_periods.append((reserved_start, end_date))
    elif previous:
        merged_reserved_periods.append((reserved_start, previous[1]))
    if (len(merged_reserved_periods) >= 1 and
            merged_reserved_periods[0][0] - start_date < duration):
        merged_reserved_periods[0] = (start_date,
                                      merged_reserved_periods[0][1])
    return merged_reserved_periods


def get_reserved_periods(resource_id, start_date, end_date, duration,
                         resource_type='host'):
    """Returns a list of reserved periods for a resource.

    The get_reserved_periods function returns a list of periods during which
    the resource passed as parameter is reserved. The duration parameter
    allows to choose the minimum length of time for a period to be
    considered free.

    :param resource_id: the resource to consider
    :param start_date: start datetime of the entire period to consider
    :param end_date: end datetime of the entire period to consider
    :param duration: minimum length of time for a period to be considered free
    :param resource_type: A type of resource to consider
    :returns: the list of reserved periods for the resource, expressed as a
              list of two-element tuples, where the first element is the start
              datetime of the reserved period and the second is
              the end datetime
    """
    capacity = 1  # The resource status is binary (free or reserved)
    quantity = 1  # One reservation per host at the same time
    if end_date - start_date < duration:
        return [(start_date, end_date)]
    events = _get_events(resource_id, start_date, end_date, resource_type)
    reserved_periods = _find_reserved_periods(events, quantity, capacity)
    return _merge_periods(reserved_periods, start_date, end_date, duration)



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from blazar.api import app as api
from blazar.api.v1 import app as v1_app
from blazar.api.v2 import app as v2_app
from blazar import tests


class FakeWSGIApp(object):
    def __init__(self, id, status_code="300 Multiple Choices"):
        self.id_version = id
        self.status_code = status_code
        self.versions = {
            "versions":
            [{"status": "CURRENT",
              "id": "v{0}".format(self.id_version),
              "links": [{
                  "href": "http://localhost/v{0}".format(self.id_version),
                  "rel": "self"}]}]}

    def __call__(self, environ, start_response):
        start_response(self.status_code, [])
        if self.status_code == "300 Multiple Choices":
            return [jsonutils.dump_as_bytes(self.versions)]
        elif self.status_code == "401 Unauthorized":
            return ['{"error": '
                    '{"message": "The request you have '
                    'made requires authentication.",'
                    '"code": 401, "title": "Unauthorized"}}']


class TestVersionDiscovery(tests.TestCase):
    scenarios = [('path_root',
                  dict(path='/')),
                 ('path_versions',
                  dict(path='/versions'))]

    def start_response(self, status, response_headers):
        pass

    def setUp(self):
        super(TestVersionDiscovery, self).setUp()
        self.start_response_mock = self.patch(self, 'start_response')
        self.v1_app = v1_app
        self.v2_app = v2_app
        self.v1_make_app = self.patch(self.v1_app, 'make_app')
        self.v1_make_app.return_value = FakeWSGIApp(1)
        self.v2_make_app = self.patch(self.v2_app, 'make_app')
        self.v2_make_app.return_value = FakeWSGIApp(2)

    def test_get_versions(self):
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': self.path}

        versions_raw = version_selector(environ, self.start_response)
        versions = jsonutils.loads(versions_raw.pop())

        self.assertEqual(2, len(versions['versions']))
        self.assertEqual("v{0}".format(self.v1_make_app().id_version),
                         versions['versions'][0]['id'])
        self.assertEqual("v{0}".format(self.v2_make_app().id_version),
                         versions['versions'][1]['id'])
        self.start_response_mock.assert_called_with(
            "300 Multiple Choices",
            [("Content-Type", "application/json")])

    def test_get_versions_only_from_one_api(self):
        self.v2_make_app.return_value = FakeWSGIApp(2, "404 Not Found")
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': self.path}

        versions_raw = version_selector(environ, self.start_response)
        versions = jsonutils.loads(versions_raw.pop())

        self.assertEqual(1, len(versions['versions']))
        self.assertEqual("v{0}".format(self.v1_make_app().id_version),
                         versions['versions'][0]['id'])
        self.start_response_mock.assert_called_with(
            "300 Multiple Choices",
            [("Content-Type", "application/json")])

    def test_no_versions_at_all(self):
        self.v1_make_app.return_value = FakeWSGIApp(1, "404 Not Found")
        self.v2_make_app.return_value = FakeWSGIApp(2, "404 Not Found")
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': self.path}

        versions_raw = version_selector(environ, self.start_response)
        self.assertEqual([], versions_raw)
        self.start_response_mock.assert_called_with("204 No Content", [])

    def test_unauthorized_token(self):
        self.v1_make_app.return_value = FakeWSGIApp(1, "401 Unauthorized")
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': self.path}

        versions_raw = version_selector(environ, self.start_response)
        self.assertEqual(['{"error": '
                          '{"message": "The request you have '
                          'made requires authentication.",'
                          '"code": 401, "title": "Unauthorized"}}'],
                         versions_raw)

        self.start_response_mock.assert_called_with(
            "401 Unauthorized",
            [("Content-Type", "application/json")])


class TestVersionSelectorApplication(tests.TestCase):
    def start_response(self, status, response_headers):
        pass

    def setUp(self):
        super(TestVersionSelectorApplication, self).setUp()
        self.v1_app = v1_app
        self.v2_app = v2_app
        self.v1_make_app = self.patch(self.v1_app, 'make_app')
        self.v1_make_app.return_value = FakeWSGIApp(1)
        self.v2_make_app = self.patch(self.v2_app, 'make_app')
        self.v2_make_app.return_value = FakeWSGIApp(2)

    def test_get_v1_app(self):
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': "/v1"}

        versions_raw = version_selector(environ, self.start_response)
        versions = jsonutils.loads(versions_raw.pop())
        self.assertEqual(self.v1_make_app().versions, versions)

    def test_get_v2_app(self):
        version_selector = api.VersionSelectorApplication()
        environ = {'PATH_INFO': "/v2"}

        versions_raw = version_selector(environ, self.start_response)
        versions = jsonutils.loads(versions_raw.pop())
        self.assertEqual(self.v2_make_app().versions, versions)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.plugins import base


class DummyVMPlugin(base.BasePlugin):
    """Plugin for VM resource that does nothing."""
    resource_type = 'virtual:instance'
    title = 'Dummy VM Plugin'
    description = 'This plugin does nothing.'

    def get(self, resource_id):
        return None

    def reserve_resource(self, reservation_id, values):
        return None

    def list_allocations(self, query, detail=False):
        """List resource allocations."""
        pass

    def query_allocations(self, resource_id_list, lease_id=None,
                          reservation_id=None):
        return None

    def allocation_candidates(self, lease_values):
        return None

    def update_reservation(self, reservation_id, values):
        return None

    def on_start(self, resource_id):
        """Dummy VM plugin does nothing."""
        return 'VM %s should be waked up this moment.' % resource_id

    def on_end(self, resource_id):
        """Dummy VM plugin does nothing."""
        return 'VM %s should be deleted this moment.' % resource_id



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from oslo_config import cfg
from oslo_context import context

CONF = cfg.CONF


class BlazarContext(context.RequestContext):

    # service_catalog is not by default read from a dict
    # when deserializing a context.
    FROM_DICT_EXTRA_KEYS = ['service_catalog']

    _context_stack = threading.local()

    def __init__(self, service_catalog=None, **kwargs):
        # NOTE(neha-alhat): During serializing/deserializing context object
        # over the RPC layer, below extra parameters which are passed by
        # `oslo.messaging` are popped as these parameters are not required.
        kwargs.pop('client_timeout', None)
        kwargs.pop('user_identity', None)

        super(BlazarContext, self).__init__(**kwargs)

        self.service_catalog = service_catalog or []

        if self.is_admin and 'admin' not in self.roles:
            self.roles.append('admin')

    def __enter__(self):
        try:
            stack = self._context_stack.stack
        except AttributeError:
            stack = []
            self._context_stack.stack = stack
        stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        res = self._context_stack.stack.pop()
        assert res is self, "self should be the top element of the stack"

    @classmethod
    def current(cls):
        try:
            return cls._context_stack.stack[-1]
        except (AttributeError, IndexError):
            raise RuntimeError("Context isn't available here")

    def to_dict(self):
        result = super(BlazarContext, self).to_dict()
        result['service_catalog'] = self.service_catalog
        return result

    @classmethod
    def admin(cls):
        try:
            cur = cls.current()
            request_id = cur.request_id
            global_request_id = cur.global_request_id
            service_catalog = cur.service_catalog
        except RuntimeError:
            request_id = global_request_id = service_catalog = None
        return cls(
            user_name=CONF.os_admin_username,
            user_domain_name=CONF.os_admin_user_domain_name,
            project_name=CONF.os_admin_project_name,
            project_domain_name=CONF.os_admin_project_domain_name,
            is_admin=True,
            service_catalog=service_catalog,
            request_id=request_id,
            global_request_id=global_request_id
        )


def current():
    return BlazarContext.current()


def admin():
    return BlazarContext.admin()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.manager import exceptions
from blazar import tests
from blazar.utils.openstack import base


class TestBaseStackUtils(tests.TestCase):

    def setUp(self):
        super(TestBaseStackUtils, self).setUp()

        self.base = base

        self.service_type = 'fake_service'
        self.url = 'http://%s-net.com'
        self.url_region_two = 'http://%s-net-two.com'

    def test_url_for_good_v3(self):
        # TODO(n.s.):Can't find v3 endpoint example. Fix it later.
        pass

    def test_url_for_good_v2_public(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"endpoints": [{"adminURL": self.url % 'admin',
                             "region": "RegionOne",
                             "internalURL": self.url % 'internal',
                             "publicURL": self.url % 'public'}],
              "type": "fake_service",
              "name": "foo"}])

        url = self.base.url_for(service_catalog, self.service_type)
        self.assertEqual(url, self.url % 'public')

    def test_url_for_good_v2_admin(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"endpoints": [{"adminURL": self.url % 'admin',
                             "region": "RegionOne",
                             "internalURL": self.url % 'internal',
                             "publicURL": self.url % 'public'}],
              "type": "fake_service",
              "name": "foo"}])

        url = self.base.url_for(service_catalog, self.service_type,
                                endpoint_interface='admin')
        self.assertEqual(url, self.url % 'admin')

    def test_url_for_no_service(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"endpoints": [{"adminURL": self.url % 'admin',
                             "region": "RegionOne",
                             "internalURL": self.url % 'internal',
                             "publicURL": self.url % 'public'}],
              "type": "foo_service",
              "name": "foo"}])

        self.assertRaises(exceptions.ServiceNotFound, self.base.url_for,
                          service_catalog, self.service_type)

    def test_url_for_no_endpoints(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"type": "fake_service",
              "name": "foo"}])

        self.assertRaises(exceptions.EndpointsNotFound, self.base.url_for,
                          service_catalog, self.service_type)

    def test_url_for_good_v2_region(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"endpoints": [{"adminURL": self.url % 'admin',
                             "region": "RegionOne",
                             "internalURL": self.url % 'internal',
                             "publicURL": self.url % 'public'},
                            {"adminURL": self.url_region_two % 'admin',
                             "region": "RegionTwo",
                             "internalURL": self.url_region_two % 'internal',
                             "publicURL": self.url_region_two % 'public'}],
              "type": "fake_service",
              "name": "foo"}])

        url = self.base.url_for(service_catalog, self.service_type,
                                os_region_name='RegionTwo')
        self.assertEqual(url, self.url_region_two % 'public')

    def test_url_for_no_endpoint_in_region(self):
        service_catalog = tests.FakeServiceCatalog(
            [{"endpoints": [{"adminURL": self.url % 'admin',
                             "region": "RegionOne",
                             "internalURL": self.url % 'internal',
                             "publicURL": self.url % 'public'}],
              "type": "fake_service",
              "name": "foo"}])

        self.assertRaises(exceptions.EndpointsNotFound, self.base.url_for,
                          service_catalog, self.service_type,
                          os_region_name='RegionTwo')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flask
from keystonemiddleware import auth_token
from oslo_config import cfg
from werkzeug import exceptions as werkzeug_exceptions

from blazar.api.v1 import app
from blazar.api.v1.leases import v1_0 as lease_api_v1_0
from blazar.api.v1.oshosts import v1_0 as host_api_v1_0
from blazar.api.v1 import utils as api_utils
from blazar import tests


class AppTestCase(tests.TestCase):
    def setUp(self):
        super(AppTestCase, self).setUp()

        self.app = app
        self.api_utils = api_utils
        self.flask = flask
        self.auth_token = auth_token

        self.render = self.patch(self.api_utils, 'render')
        self.fake_ff = self.patch(self.auth_token, 'filter_factory')

        self.ex = werkzeug_exceptions.HTTPException()
        self.ex.code = 1313
        self.ex.description = "my favourite error"

    def test_make_json_error_proper(self):
        self.app.make_json_error(self.ex)
        self.render.assert_called_once_with(
            {'error': 1313,
             'error_message': 'my favourite error'}, status=1313)

    def test_make_json_error_wrong(self):
        self.app.make_json_error('wrong')
        self.render.assert_called_once_with(
            {'error': 500,
             'error_message': 'wrong'}, status=500)

    def test_version_list(self):
        a = flask.Flask('dummy')
        a.config['TESTING'] = True
        with a.test_request_context():
            self.app.version_list()
            self.render.assert_called_once_with({
                "versions": [
                    {"id": "v1.0",
                     "status": "CURRENT",
                     'min_version': '1.0', 'max_version': '1.0',
                     "links": [{"href": "{0}v1".format(flask.request.host_url),
                                "rel": "self"}]
                     },
                ]},
                status="300 Multiple Choices")

    def test_make_app(self):
        fake_app = self.patch(self.flask, 'Flask')
        self.app.make_app()
        self.fake_ff.assert_called_once_with(fake_app().config)


class AppTestCaseForHostsPlugin(tests.TestCase):

    def setUp(self):
        super(AppTestCaseForHostsPlugin, self).setUp()

        cfg.CONF.set_override('plugins', ['physical.host.plugin'], 'manager')
        self.app = app
        self.host_api_v1_0 = host_api_v1_0
        self.lease_api_v1_0 = lease_api_v1_0
        self.flask = flask
        self.fake_blueprint = self.patch(self.flask.Flask,
                                         'register_blueprint')

    def test_make_app_with_host_plugin(self):
        self.app.make_app()
        self.fake_blueprint.assert_any_call(self.lease_api_v1_0.rest,
                                            url_prefix='/v1')
        self.fake_blueprint.assert_any_call(self.host_api_v1_0.rest,
                                            url_prefix='/v1/os-hosts')



# Copyright (c) 2022 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import json
import requests
from urllib.parse import urljoin
from urllib.parse import urlparse

from blazar.enforcement.exceptions import ExternalServiceFilterException
from blazar.enforcement.filters import base_filter
from blazar.exceptions import BlazarException
from blazar.i18n import _

from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class ISODateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


GENERIC_DENY_MSG = 'External service enforcement filter denied the request.'


class ExternalServiceMisconfigured(BlazarException):
    msg_fmt = _('%(message)s')


class ExternalServiceFilter(base_filter.BaseFilter):

    enforcement_opts = [
        cfg.StrOpt(
            'external_service_base_endpoint',
            default=None,
            help='The URL of the external service API.'),
        cfg.StrOpt(
            'external_service_check_create_endpoint',
            default=None,
            help='Overrides check-create endpoint with another URL.'),
        cfg.StrOpt(
            'external_service_check_update_endpoint',
            default=None,
            help='Overrides check-update endpoint with another URL.'),
        cfg.StrOpt(
            'external_service_on_end_endpoint',
            default=None,
            help='Overrides on-end endpoint with another URL.'),
        cfg.StrOpt(
            'external_service_token',
            default="",
            help='Token used for authentication with the external service.')
    ]

    def __init__(self, conf=None):
        super(ExternalServiceFilter, self).__init__(conf=conf)

        self._validate_url(conf.enforcement.external_service_base_endpoint)
        self.base_endpoint = conf.enforcement.external_service_base_endpoint

        self.check_create_endpoint = self._construct_url(
            "check-create",
            conf.enforcement.external_service_check_create_endpoint)
        self.check_update_endpoint = self._construct_url(
            "check-update",
            conf.enforcement.external_service_check_update_endpoint)
        self.on_end_endpoint = self._construct_url(
            "on-end",
            conf.enforcement.external_service_on_end_endpoint)

        endpoints = (
            self.check_create_endpoint,
            self.check_update_endpoint,
            self.on_end_endpoint,
        )

        if all(x is None for x in endpoints):
            raise ExternalServiceMisconfigured(
                message=_("ExternalService has no endpoints set."))

        self.token = conf.enforcement.external_service_token

    @staticmethod
    def _validate_url(url):
        if url is None:
            return
        parsed_url = urlparse(url)
        if parsed_url.scheme not in ("http", "https"):
            raise ExternalServiceMisconfigured(
                message=_("ExternalService URL scheme must be http(s): "
                          "%s") % url)
        if parsed_url.netloc == '':
            raise ExternalServiceMisconfigured(
                message=_("ExternalService URL must have netloc: "
                          "%s") % url)

    def _construct_url(self, method, replacement_url):
        if replacement_url is None:
            if self.base_endpoint is None:
                return None
            return urljoin(self.base_endpoint, method)

        self._validate_url(replacement_url)
        return replacement_url

    def _get_headers(self):
        headers = {'Content-Type': 'application/json'}

        if self.token:
            headers['X-Auth-Token'] = self.token

        return headers

    def _post(self, url, body):
        body = json.dumps(body, cls=ISODateTimeEncoder)
        res = requests.post(url, headers=self._get_headers(), data=body)

        if res.status_code == 204:
            return True
        elif res.status_code == 403:
            try:
                message = res.json()['message']
            except (requests.JSONDecodeError, KeyError):
                # NOTE(yoctozepto): It is more secure not to send the actual
                # response to the end user as it may leak something.
                # Instead, we log it for debugging.
                LOG.debug("The External Service API returned a malformed "
                          "response (403): %s", res.content)
                message = GENERIC_DENY_MSG
        else:
            # NOTE(yoctozepto): It is more secure not to send the actual
            # response to the end user as it may leak something.
            # Instead, we log it for debugging.
            LOG.debug("The External Service API returned a malformed "
                      "response (%d): %s", res.status_code, res.content)
            message = GENERIC_DENY_MSG
        raise ExternalServiceFilterException(message=message)

    def check_create(self, context, lease_values):
        if self.check_create_endpoint:
            self._post(self.check_create_endpoint, dict(
                context=context, lease=lease_values))

    def check_update(self, context, current_lease_values, new_lease_values):
        if self.check_update_endpoint:
            self._post(self.check_update_endpoint, dict(
                context=context, current_lease=current_lease_values,
                lease=new_lease_values))

    def on_end(self, context, lease_values):
        if self.on_end_endpoint:
            self._post(self.on_end_endpoint, dict(
                context=context, lease=lease_values))



# Copyright (c) 2017 NTT Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from oslo_config import cfg
from oslo_log import log as logging

from blazar.api import app as wsgi_app
from blazar.api.v2 import app as v2_app
from blazar.utils import service as service_utils


CONF = cfg.CONF
CONF.import_opt('enable_v1_api', 'blazar.config')
logging.register_options(CONF)

CONFIG_FILES = ['blazar.conf']


def _get_config_files(env=None):
    if env is None:
        env = os.environ
    dirname = env.get('OS_BLAZAR_CONFIG_DIR', '/etc/blazar').strip()
    return [os.path.join(dirname, config_file)
            for config_file in CONFIG_FILES]


def init_app():
    config_files = _get_config_files()
    CONF([], project='blazar', prog='blazar-api',
         default_config_files=config_files)
    service_utils.prepare_service()
    if not CONF.enable_v1_api:
        app = v2_app.make_app()
    else:
        app = wsgi_app.VersionSelectorApplication()

    return app



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from blazar.policies import base

POLICY_ROOT = 'blazar:oshosts:%s'

oshosts_policies = [
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'get',
        check_str=base.RULE_ADMIN,
        description='Policy rule for List/Show Host(s) API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts',
                'method': 'GET'
            },
            {
                'path': '/{api_version}/os-hosts/{host_id}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'post',
        check_str=base.RULE_ADMIN,
        description='Policy rule for Create Host API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'put',
        check_str=base.RULE_ADMIN,
        description='Policy rule for Update Host API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts/{host_id}',
                'method': 'PUT'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'delete',
        check_str=base.RULE_ADMIN,
        description='Policy rule for Delete Host API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts/{host_id}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'get_allocations',
        check_str=base.RULE_ADMIN,
        description='Policy rule for List/Get Host(s) Allocations API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts/allocations',
                'method': 'GET'
            },
            {
                'path': '/{api_version}/os-hosts/{host_id}/allocation',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'get_resource_properties',
        check_str=base.RULE_ADMIN,
        description='Policy rule for Resource Properties API.',
        operations=[
            {
                'path': '/{api_version}/os-hosts/resource_properties',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'update_resource_properties',
        check_str=base.RULE_ADMIN,
        description='Policy rule for Resource Properties API.',
        operations=[
            {
                'path': ('/{api_version}/os-hosts/resource_properties/'
                         '{property_name}'),
                'method': 'PATCH'
            }
        ]
    ),
]


def list_rules():
    return oshosts_policies



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar import exceptions
from blazar.i18n import _


class NotImplemented(exceptions.BlazarException):
    code = 409
    msg_fmt = '%(error)s'


class ReservationTypeConflict(exceptions.NotFound):
    code = 409
    msg_fmt = _("Conflicting reservation types found")


class NoFreePool(exceptions.NotFound):
    msg_fmt = _("No Freepool found")


class HostNotInFreePool(exceptions.NotFound):
    msg_fmt = _("Host %(host)s not in freepool '%(freepool_name)s'")


class CantRemoveHost(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't remove host(s) %(host)s from Aggregate %(pool)s")


class CantDeleteHost(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't delete host %(host)s. %(msg)s")


class CantAddHost(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't add host(s) %(host)s to Aggregate %(pool)s")


class AggregateHaveHost(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't delete Aggregate '%(name)s', "
                "host(s) attached to it : %(hosts)s")


class AggregateAlreadyHasHost(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Conflict while adding host %(host)s to aggregate %(pool)s: "
                "%(nova_exception)s")


class AggregateNotFound(exceptions.NotFound):
    msg_fmt = _("Aggregate '%(pool)s' not found!")


class HostNotFound(exceptions.NotFound):
    msg_fmt = _("Host '%(host)s' not found!")


class InvalidHost(exceptions.NotAuthorized):
    msg_fmt = _("Invalid values for host %(host)s")


class MultipleHostsFound(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Multiple Hosts found for pattern '%(host)s'")


class HostHavingServers(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Servers [%(servers)s] found for host %(host)s")


class PluginConfigurationError(exceptions.BlazarException):
    msg_fmt = _("Plugin Configuration error : %(error)s")


class EventError(exceptions.BlazarException):
    msg_fmt = '%(error)s'


class InvalidDate(exceptions.BlazarException):
    code = 400
    msg_fmt = _(
        '%(date)s is an invalid date. Required format: %(date_format)s')


class UnsupportedResourceType(exceptions.BlazarException):
    msg_fmt = _("The %(resource_type)s resource type is not supported")


class LeaseNameAlreadyExists(exceptions.BlazarException):
    code = 409
    msg_fmt = _("The lease with name: %(name)s already exists")


class MissingTrustId(exceptions.BlazarException):
    msg_fmt = _("A trust id is required")


class NovaClientError(exceptions.BlazarException):
    msg_fmt = _("Failed to create Nova resources for the reservation")


class ServerDeletionTimeout(exceptions.BlazarException):
    msg_fmt = _("Timed out waiting for server deletion")


# oshost plugin related exceptions

class CantAddExtraCapability(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't add extracapabilities %(keys)s to Host %(host)s")


class EndpointsNotFound(exceptions.NotFound):
    code = 404
    msg_fmt = _("No endpoints for %(service)s")


class ServiceNotFound(exceptions.NotFound):
    code = 404
    msg_fmt = _("Service %(service)s not found")


class WrongClientVersion(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Unfortunately you use wrong client version")


class NoManagementUrl(exceptions.NotFound):
    code = 404
    msg_fmt = _("You haven't management url for service")


class HypervisorNotFound(exceptions.BlazarException):
    msg_fmt = _("Aggregate '%(pool)s' not found!")


class NotEnoughHostsAvailable(exceptions.BlazarException):
    msg_fmt = _("Not enough hosts available")


class MalformedParameter(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Malformed parameter %(param)s")


class MalformedRequirements(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Malformed requirements %(rqrms)s")


class MissingParameter(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Missing parameter %(param)s")


class ExtraCapabilityTooLong(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Extra capability key too long")


class InvalidState(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Invalid State %(state)s for %(id)s")


class InvalidStateUpdate(InvalidState):
    msg_fmt = _("Unable to update ID %(id)s state with %(action)s:%(status)s")


class ProjectIdNotFound(exceptions.BlazarException):
    msg_fmt = _("No project_id found in current context")


class InvalidRange(exceptions.BlazarException):
    code = 400
    msg_fmt = _('Invalid values for min/max of hosts. '
                'Max must be equal to or larger than min.')


class CantUpdateParameter(exceptions.BlazarException):
    code = 409
    msg_fmt = _("%(param)s cannot be updated")


class InvalidPeriod(exceptions.BlazarException):
    code = 400
    msg_fmt = _('The end_date must be later than the start_date.')


# floating ip plugin related exceptions

class FloatingIPNotFound(exceptions.NotFound):
    msg_fmt = _("Floating IP %(floatingip)s not found.")


class CantDeleteFloatingIP(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Can't delete floating IP %(floatingip)s. %(msg)s")


class InvalidIPFormat(exceptions.InvalidInput):
    msg_fmt = _("IP address %(ip)s is invalid form.")


class TooLongFloatingIPs(exceptions.InvalidInput):
    msg_fmt = _("Invalid values for required_floatingips and amount. "
                "The amount must be equal to or longer than length of "
                "required_floatingips.")


class NotEnoughFloatingIPAvailable(exceptions.InvalidInput):
    msg_fmt = _("Not enough floating IPs available")


class CantUpdateFloatingIPReservation(exceptions.BlazarException):
    code = 400
    msg_fmt = _("Floating IP reservation cannot be updated with requested "
                "parameters. %(msg)s")


class NeutronClientError(exceptions.BlazarException):
    msg_fmt = _("Failed to create Neutron resources for the reservation")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
from unittest import mock

from novaclient.v2 import flavors

from blazar import context
from blazar.db.sqlalchemy import api as db_api
from blazar.db import utils as db_utils
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins.flavor import flavor_plugin
from blazar.plugins.oshosts import host_plugin
from blazar import tests
from blazar.tests.db.sqlalchemy import test_sqlalchemy_api as fake
from blazar.utils.openstack import nova
from blazar.utils.openstack import placement


class TestFlavorPlugin(tests.DBTestCase):
    def _create_fake_host(self):
        host_values = fake._get_fake_host_values(id=123)
        host_values["reservable"] = 1
        db_api.host_create(host_values)

    def test_get(self):
        plugin = flavor_plugin.FlavorPlugin()
        resource_id = '123'
        self._create_fake_host()

        result = plugin.get(resource_id)

        self.assertEqual(resource_id, result['id'])

    def test_list_allocations(self):
        plugin = flavor_plugin.FlavorPlugin()

        result = plugin.list_allocations({'lease_id': '2001'})

        self.assertEqual(0, len(result))

    @mock.patch.object(host_plugin.PhysicalHostPlugin, 'query_allocations')
    def test_query_allocations(self, mock_query):
        plugin = flavor_plugin.FlavorPlugin()
        mock_query.return_value = "fake"

        result = plugin.query_allocations(['123'], lease_id='2001')

        self.assertEqual("fake", result)
        mock_query.assert_called_once_with(['123'], '2001', None)

    @mock.patch.object(flavor_plugin.FlavorPlugin, '_get_flavor_details')
    def test_allocation_candidates(self, mock_get_flavor):
        self._create_fake_host()
        fake_inventory_values = {
            'computehost_id': 123,
            'resource_class': 'PCPU',
            'total': 10,
            'reserved': 2,
            'min_unit': 1,
            'max_unit': 10,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)
        plugin = flavor_plugin.FlavorPlugin()
        reservation = {
            'flavor_id': "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            'amount': 4,
            'affinity': None,
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
        }
        mock_get_flavor.return_value = ({"PCPU": 2}, {},
                                        {"flavor_id": "fake"})

        result = plugin.allocation_candidates(reservation)

        self.assertEqual(4, len(result))
        mock_get_flavor.assert_called_once_with(
            "34eb7166-0e9b-432c-96fd-dff37f22e36e")

    @mock.patch.object(flavor_plugin.FlavorPlugin, '_get_flavor_details')
    def test_allocation_candidates_fails_no_space(self, mock_get_flavor):
        self._create_fake_host()
        fake_inventory_values = {
            'computehost_id': 123,
            'resource_class': 'PCPU',
            'total': 10,
            'reserved': 2,
            'min_unit': 1,
            'max_unit': 10,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)
        plugin = flavor_plugin.FlavorPlugin()
        reservation = {
            'flavor_id': "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            'amount': 5,
            'affinity': None,
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
        }
        mock_get_flavor.return_value = ({"PCPU": 2}, {},
                                        {"flavor_id": "fake"})

        self.assertRaises(mgr_exceptions.NotEnoughHostsAvailable,
                          plugin.allocation_candidates,
                          reservation)

    @mock.patch.object(flavor_plugin.FlavorPlugin, '_create_resources')
    @mock.patch.object(flavor_plugin.FlavorPlugin, '_get_flavor_details')
    def test_allocation_candidates_avoids_reservations(self, mock_get_flavor,
                                                       mock_create):
        self._create_fake_host()
        fake_inventory_values = {
            'computehost_id': 123,
            'resource_class': 'PCPU',
            'total': 10,
            'reserved': 2,
            'min_unit': 1,
            'max_unit': 10,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)
        plugin = flavor_plugin.FlavorPlugin()
        new_reservation = {
            'flavor_id': "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            'amount': 3,
            'affinity': None,
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
        }
        fake_flavor = {
            "disk": 0,  # GiB
            "OS-FLV-EXT-DATA:ephemeral": 0,  # GiB
            "id": "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            "name": "test1",
            "ram": 0,  # MB
            "swap": 0,
            "vcpus": 2,
            "extra_specs": {'hw:cpu_policy': 'dedicated'}
        }
        mock_get_flavor.return_value = ({"PCPU": 2}, {},
                                        fake_flavor)
        old_reservation = new_reservation.copy()
        old_reservation['amount'] = 2
        fake_phys_reservation = new_reservation.copy()
        fake_phys_reservation['id'] = 345
        fake_start_event = {
            'id': 123,
            'lease_id': 1234,
            'event_type': "start_lease",
            'time': datetime.datetime(2030, 1, 1, 8, 00),
            'status': "fake",
        }
        fake_lease = {
            'id': 1234,
            'name': "fakelease",
            'user_id': 'fake',
            'project_id': 'fake',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00),
            'trust': 'trust',
            'reservations': [fake_phys_reservation],
            'events': [fake_start_event]
        }
        db_api.lease_create(fake_lease)
        mock_create.return_value = ("flavor_id", "aggregate_id")
        # create a reservation to avoid
        plugin.reserve_resource("345", old_reservation)

        # Host as 10 PCPUs, 2 are reserved, leaving 8 PCPUs available
        # Old reservation is for 2 flavors needing 2 each, so 4 left
        # So there should be space for 2 lots of 2 PCPUs
        new_reservation['amount'] = 2
        result = plugin.allocation_candidates(new_reservation)
        self.assertEqual(2, len(result))

        # there should not be space for 3 lots of 2 PCPUs
        new_reservation['amount'] = 3
        self.assertRaises(mgr_exceptions.NotEnoughHostsAvailable,
                          plugin.allocation_candidates,
                          new_reservation)

    @mock.patch.object(flavors.FlavorManager, 'get')
    def test__get_flavor_details(self, mock_get):
        plugin = flavor_plugin.FlavorPlugin()
        mock_flavor = mock.Mock()
        mock_get.return_value = mock_flavor
        fake_flavor = {
            "disk": 10,  # GiB
            "OS-FLV-EXT-DATA:ephemeral": 0,  # GiB
            "id": "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            "name": "test1",
            "ram": 1024,  # MB
            "swap": 0,
            "vcpus": 1,
        }
        mock_flavor.to_dict.return_value = fake_flavor
        fake_extra_specs = {
            'hw:cpu_policy': 'dedicated',
            'trait:HW_CPU_X86_AVX': 'required',
            'resources:VGPU': '1',
        }
        mock_flavor.get_keys.return_value = fake_extra_specs

        resource_request, resource_traits, source_flavor = \
            plugin._get_flavor_details("34eb7166-0e9b-432c-96fd-dff37f22e36e")

        self.assertDictEqual({
            'DISK_GB': 10, 'MEMORY_MB': 1024, 'PCPU': 1, 'VCPU': 0, 'VGPU': 1
        }, resource_request)
        self.assertDictEqual({'HW_CPU_X86_AVX': 'required'}, resource_traits)
        expected = fake_flavor.copy()
        expected["extra_specs"] = fake_extra_specs
        self.assertDictEqual(expected, source_flavor)

    @mock.patch.object(flavor_plugin.FlavorPlugin, '_create_resources')
    @mock.patch.object(flavor_plugin.FlavorPlugin, '_pick_hosts')
    def test_reserve_resource(self, mock_pick, mock_create):
        plugin = flavor_plugin.FlavorPlugin()
        reservation = {
            'flavor_id': "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            'amount': 5,
            'affinity': None,
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
        }
        fake_flavor = {
            "disk": 10,  # GiB
            "OS-FLV-EXT-DATA:ephemeral": 0,  # GiB
            "id": "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            "name": "test1",
            "ram": 1024,  # MB
            "swap": 0,
            "vcpus": 1,
        }
        mock_pick.return_value = (['123', '124'], fake_flavor)
        mock_create.return_value = ("flavor_id", "aggregate_id")

        id = plugin.reserve_resource("345", reservation)

        allocations = db_api.host_allocation_get_all_by_values(
            reservation_id="345")
        self.assertEqual(2, len(allocations))
        reservation = db_api.instance_reservation_get(id)
        self.assertEqual("345", reservation["reservation_id"])
        self.assertEqual("flavor_id", reservation["flavor_id"])
        self.assertEqual("aggregate_id", reservation["aggregate_id"])

    @mock.patch.object(nova.ReservationPool, 'create')
    @mock.patch.object(context.BlazarContext, 'current')
    @mock.patch.object(placement.BlazarPlacementClient,
                       'create_reservation_class')
    @mock.patch.object(flavor_plugin.FlavorPlugin, '_create_flavor')
    def test_create_resources(self, mock_create_flavor,
                              mock_reservation_create,
                              mock_current, mock_pool_create):
        plugin = flavor_plugin.FlavorPlugin()
        fake_flavor = {
            "disk": 10,  # GiB
            "OS-FLV-EXT-DATA:ephemeral": 100,  # GiB
            "id": "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            "name": "test1",
            "ram": 1024,  # MB
            "swap": 0,
            "vcpus": 2,
            "extra_specs": {'hw:cpu_policy': 'dedicated'}
        }
        fake_reservation = {
            'reservation_id': "12345",
            'vcpus': fake_flavor["vcpus"],
            'memory_mb': fake_flavor["ram"],
            'disk_gb': fake_flavor["disk"],
            'amount': 2,
            'affinity': None,
            'resource_properties': json.dumps(fake_flavor)
        }
        mock_context = mock.Mock()
        mock_context.project_id = "fake-project-id"
        mock_current.return_value = mock_context
        mock_aggregate = mock.Mock()
        mock_aggregate.id = "aggregate_id"
        mock_pool_create.return_value = mock_aggregate
        mock_flavor = mock.Mock()
        mock_flavor.id = "flavor_id"
        mock_create_flavor.return_value = mock_flavor

        fid, aid = plugin._create_resources(fake_reservation)

        self.assertEqual(fid, "flavor_id")
        self.assertEqual(aid, "aggregate_id")
        mock_create_flavor.assert_called_once_with(fake_reservation)
        mock_reservation_create.assert_called_once_with("12345")
        mock_current.assert_called_once_with()
        mock_pool_create.assert_called_once_with(
            name="12345",
            metadata={'reservation': '12345',
                      'filter_tenant_id': 'fake-project-id'}
        ),

    @mock.patch.object(flavors.FlavorManager, 'create')
    def test_create_flavor(self, mock_create):
        plugin = flavor_plugin.FlavorPlugin()
        fake_flavor = {
            "disk": 10,  # GiB
            "OS-FLV-EXT-DATA:ephemeral": 100,  # GiB
            "id": "34eb7166-0e9b-432c-96fd-dff37f22e36e",
            "name": "test1",
            "ram": 1024,  # MB
            "swap": 0,
            "vcpus": 2,
            "extra_specs": {'hw:cpu_policy': 'dedicated'}
        }
        fake_reservation = {
            'reservation_id': "12345",
            'vcpus': fake_flavor["vcpus"],
            'memory_mb': fake_flavor["ram"],
            'disk_gb': fake_flavor["disk"],
            'amount': 2,
            'affinity': None,
            'resource_properties': json.dumps(fake_flavor)
        }
        mock_flavor = mock.Mock()
        mock_create.return_value = mock_flavor

        plugin._create_flavor(fake_reservation)

        mock_flavor.set_keys.assert_called_once_with({
            'hw:cpu_policy': 'dedicated',
            'aggregate_instance_extra_specs:reservation': '12345',
            'resources:CUSTOM_RESERVATION_12345': '1',
        })
        mock_create.assert_called_once_with(
            flavorid='12345', name='reservation:12345', vcpus=2, ram=1024,
            disk=10, is_public=False)

    def test__query_available_hosts(self):
        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_id')
        get_reservations.return_value = []
        plugin = flavor_plugin.FlavorPlugin()

        # Check that we can fit 4 VCPU resource requests
        self._create_fake_host()
        fake_inventory_values = {
            'computehost_id': 123,
            'resource_class': 'VCPU',
            'total': 4,
            'reserved': 0,
            'min_unit': 1,
            'max_unit': 4,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)

        query_params = {
            'start_date': datetime.datetime(2020, 7, 7, 18, 0),
            'end_date': datetime.datetime(2020, 7, 7, 19, 0),
            'resource_request': {
                'VCPU': 1,
            },
            'resource_traits': {}
        }
        ret = plugin._query_available_hosts(**query_params)
        self.assertEqual(4, len(ret))

        # Only 2 * 1024 MB requests fit when we add a MEMORY_MB inventory
        fake_inventory_values = {
            'computehost_id': 123,
            'resource_class': 'MEMORY_MB',
            'total': 2048,
            'reserved': 0,
            'min_unit': 1,
            'max_unit': 2048,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)

        query_params = {
            'start_date': datetime.datetime(2020, 7, 7, 18, 0),
            'end_date': datetime.datetime(2020, 7, 7, 19, 0),
            'resource_request': {
                'VCPU': 1,
                'MEMORY_MB': 1024
            },
            'resource_traits': {}
        }
        ret = plugin._query_available_hosts(**query_params)
        self.assertEqual(2, len(ret))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils
import sqlalchemy as sa


class JsonEncoded(sa.TypeDecorator):
    """Represents an immutable structure as a json-encoded string."""

    impl = sa.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dump_as_bytes(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value



# Copyright (c) 2019 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from blazar import manager
from blazar.utils import service

CONF = cfg.CONF
CONF.import_opt('rpc_topic', 'blazar.manager.service', 'manager')


class ManagerRPCAPI(service.RPCClient):
    """Client side for the Manager RPC API.

    Used from other services to communicate with blazar-manager service.
    """
    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        """Initiate RPC API client with needed topic and RPC version."""
        super(ManagerRPCAPI, self).__init__(manager.get_target())

    def get_floatingip(self, floatingip_id):
        """Get detailed info about a floatingip."""
        return self.call('virtual:floatingip:get_floatingip',
                         fip_id=floatingip_id)

    def list_floatingips(self):
        """List all floatingips."""
        return self.call('virtual:floatingip:list_floatingip')

    def create_floatingip(self, floatingip_values):
        """Create floatingip with specified parameters."""
        return self.call('virtual:floatingip:create_floatingip',
                         values=floatingip_values)

    def delete_floatingip(self, floatingip_id):
        """Delete specified floatingip."""
        return self.call('virtual:floatingip:delete_floatingip',
                         fip_id=floatingip_id)



# Copyright (c) 2020 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import ddt

from blazar import context
from blazar import enforcement
from blazar.enforcement import filters
from blazar import exceptions
from blazar.manager import service
from blazar import tests

from oslo_config import cfg


def get_fake_host(host_id):
    return {
        'id': host_id,
        'hypervisor_hostname': 'hypvsr1',
        'service_name': 'compute1',
        'vcpus': 4,
        'cpu_info': 'foo',
        'hypervisor_type': 'xen',
        'hypervisor_version': 1,
        'memory_mb': 8192,
        'local_gb': 10,
    }


def get_fake_lease(**kwargs):
    fake_lease = {
        'id': '1',
        'name': 'lease_test',
        'start_date': datetime.datetime.utcnow().strftime(
            service.LEASE_DATE_FORMAT),
        'end_date': (
            datetime.datetime.utcnow() + datetime.timedelta(days=1)).strftime(
                service.LEASE_DATE_FORMAT),
        'user_id': '111',
        'project_id': '222',
        'reservations': [{'resource_id': '1234',
                          'resource_type': 'virtual:instance'}],
        'events': [],
        'before_end_date': '2014-02-01 10:37',
        'action': None,
        'status': None,
        'status_reason': None,
        'trust_id': 'exxee111qwwwwe'}

    if kwargs:
        fake_lease.update(kwargs)

    return fake_lease


def get_lease_rsv_allocs():
    allocation_candidates = {'virtual:instance': [get_fake_host('1')]}
    lease_values = get_fake_lease()
    reservations = list(lease_values['reservations'])

    del lease_values['reservations']

    return lease_values, reservations, allocation_candidates


class FakeFilter(filters.base_filter.BaseFilter):

    enforcement_opts = [
        cfg.IntOpt('fake_opt', default=1, help='This is a fake config.'),
    ]

    def __init__(self, conf=None):
        super(FakeFilter, self).__init__(conf=conf)

    def check_create(self, context, lease_values):
        pass

    def check_update(self, context, current_lease_values, new_lease_values):
        pass

    def on_end(self, context, lease_values):
        pass


@ddt.ddt
class EnforcementTestCase(tests.TestCase):
    def setUp(self):
        super(EnforcementTestCase, self).setUp()

        self.cfg = cfg
        self.region = 'RegionOne'
        filters.FakeFilter = FakeFilter
        filters.all_filters = ['FakeFilter']

        self.enforcement = enforcement.UsageEnforcement()

        cfg.CONF.set_override(
            'enabled_filters', filters.all_filters, group='enforcement')
        cfg.CONF.set_override('os_region_name', self.region)

        self.enforcement.load_filters()
        self.fake_service_catalog = tests.FakeServiceCatalog([
            dict(
                type='identity', endpoints=[
                    dict(
                        interface='internal', region=self.region,
                        url='https://fakeauth.com')
                ]
            )
        ])

        self.ctx = context.BlazarContext(
            user_id='111', project_id='222',
            service_catalog=self.fake_service_catalog)
        self.set_context(self.ctx)

        self.fake_host_id = '1'
        self.fake_host = {
            'id': self.fake_host_id,
            'hypervisor_hostname': 'hypvsr1',
            'service_name': 'compute1',
            'vcpus': 4,
            'cpu_info': 'foo',
            'hypervisor_type': 'xen',
            'hypervisor_version': 1,
            'memory_mb': 8192,
            'local_gb': 10,
        }
        self.addCleanup(self.cfg.CONF.clear_override, 'enabled_filters',
                        group='enforcement')
        self.addCleanup(self.cfg.CONF.clear_override, 'os_region_name')

    def tearDown(self):
        super(EnforcementTestCase, self).tearDown()

    def get_formatted_lease(self, lease_values, rsv, allocs):
        expected_lease = lease_values.copy()

        if rsv:
            expected_lease['reservations'] = rsv
        for res in expected_lease['reservations']:
            res['allocations'] = allocs[res['resource_type']]

        return expected_lease

    def test_load_filters(self):
        self.assertEqual(len(self.enforcement.enabled_filters), 1)

        fake_filter = self.enforcement.enabled_filters.pop()

        self.assertIsInstance(fake_filter, FakeFilter)
        self.assertEqual(fake_filter.conf.enforcement.fake_opt, 1)

    def test_format_context(self):

        formatted_context = self.enforcement.format_context(
            context.current(), get_fake_lease())

        expected = dict(user_id='111', project_id='222',
                        region_name=self.region,
                        auth_url='https://fakeauth.com')

        self.assertDictEqual(expected, formatted_context)

    def test_format_lease(self):
        lease_values, rsv, allocs = get_lease_rsv_allocs()

        formatted_lease = self.enforcement.format_lease(lease_values, rsv,
                                                        allocs)

        expected_lease = self.get_formatted_lease(lease_values, rsv, allocs)

        self.assertDictEqual(expected_lease, formatted_lease)

    def test_check_create(self):
        lease_values, rsv, allocs = get_lease_rsv_allocs()
        ctx = context.current()

        check_create = self.patch(self.enforcement.enabled_filters[0],
                                  'check_create')

        self.enforcement.check_create(ctx, lease_values, rsv, allocs)

        formatted_lease = self.enforcement.format_lease(lease_values, rsv,
                                                        allocs)
        formatted_context = self.enforcement.format_context(ctx, lease_values)

        check_create.assert_called_once_with(formatted_context,
                                             formatted_lease)

        expected_context = dict(user_id='111', project_id='222',
                                region_name=self.region,
                                auth_url='https://fakeauth.com')

        expected_lease = self.get_formatted_lease(lease_values, rsv, allocs)

        self.assertDictEqual(expected_context, formatted_context)
        self.assertDictEqual(expected_lease, formatted_lease)

    def test_check_create_with_exception(self):
        lease_values, rsv, allocs = get_lease_rsv_allocs()
        ctx = context.current()

        check_create = self.patch(self.enforcement.enabled_filters[0],
                                  'check_create')

        check_create.side_effect = exceptions.BlazarException

        self.assertRaises(exceptions.BlazarException,
                          self.enforcement.check_create,
                          context=ctx, lease_values=lease_values,
                          reservations=rsv, allocations=allocs)

    def test_check_update(self):
        lease, rsv, allocs = get_lease_rsv_allocs()

        new_lease_values = get_fake_lease(end_date='2014-02-07 13:37')
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        check_update = self.patch(self.enforcement.enabled_filters[0],
                                  'check_update')

        self.enforcement.check_update(
            ctx, lease, new_lease_values, allocs, allocation_candidates,
            rsv, new_reservations)

        formatted_context = self.enforcement.format_context(ctx, lease)
        formatted_lease = self.enforcement.format_lease(lease, rsv, allocs)
        new_formatted_lease = self.enforcement.format_lease(
            new_lease_values, new_reservations, allocation_candidates)

        expected_context = dict(user_id='111', project_id='222',
                                region_name=self.region,
                                auth_url='https://fakeauth.com')

        expected_lease = self.get_formatted_lease(lease, rsv, allocs)
        expected_new_lease = self.get_formatted_lease(
            new_lease_values, new_reservations, allocation_candidates)

        check_update.assert_called_once_with(
            formatted_context, formatted_lease, new_formatted_lease)

        self.assertDictEqual(expected_context, formatted_context)
        self.assertDictEqual(expected_lease, formatted_lease)
        self.assertDictEqual(expected_new_lease, new_formatted_lease)

    def test_check_update_with_exception(self):
        lease, rsv, allocs = get_lease_rsv_allocs()

        new_lease_values = get_fake_lease(end_date='2014-02-07 13:37')
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        check_update = self.patch(self.enforcement.enabled_filters[0],
                                  'check_update')
        check_update.side_effect = exceptions.BlazarException

        self.assertRaises(
            exceptions.BlazarException, self.enforcement.check_update,
            context=ctx, current_lease=lease, new_lease=new_lease_values,
            current_allocations=allocs, new_allocations=allocation_candidates,
            current_reservations=rsv, new_reservations=new_reservations)

    def test_on_end(self):
        allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease()
        ctx = context.current()

        on_end = self.patch(self.enforcement.enabled_filters[0], 'on_end')

        self.enforcement.on_end(ctx, lease, allocations)

        formatted_context = self.enforcement.format_context(ctx, lease)
        formatted_lease = self.enforcement.format_lease(
            lease, lease['reservations'], allocations)

        on_end.assert_called_once_with(formatted_context, formatted_lease)

        expected_context = dict(user_id='111', project_id='222',
                                region_name=self.region,
                                auth_url='https://fakeauth.com')

        expected_lease = self.get_formatted_lease(lease, None, allocations)

        self.assertDictEqual(expected_context, formatted_context)
        self.assertDictEqual(expected_lease, formatted_lease)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar import manager
from blazar.utils import service


class ManagerRPCAPI(service.RPCClient):
    """Client side for the Manager RPC API.

    Used from other services to communicate with blazar-manager service.
    """
    def __init__(self):
        """Initiate RPC API client with needed topic and RPC version."""
        super(ManagerRPCAPI, self).__init__(manager.get_target())

    def get_lease(self, lease_id):
        """Get detailed info about some lease."""
        return self.call('get_lease', lease_id=lease_id)

    def list_leases(self, project_id=None, query=None):
        """List all leases."""
        return self.call('list_leases', project_id=project_id, query=query)

    def create_lease(self, lease_values):
        """Create lease with specified parameters."""
        return self.call('create_lease', lease_values=lease_values)

    def update_lease(self, lease_id, values):
        """Update lease with passes values dictionary."""
        return self.call('update_lease', lease_id=lease_id, values=values)

    def delete_lease(self, lease_id):
        """Delete specified lease."""
        return self.call('delete_lease', lease_id=lease_id)



# Copyright 2017 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""instance-reservation-table

Revision ID: 8805be233864
Revises: 7f1a7bbb2cd2
Create Date: 2017-06-20 07:31:17.379252

"""

# revision identifiers, used by Alembic.
revision = '8805be233864'
down_revision = '7f1a7bbb2cd2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('instance_reservations',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('reservation_id',
                              sa.String(length=36), nullable=True),
                    sa.Column('vcpus', sa.Integer(), nullable=False),
                    sa.Column('memory_mb', sa.Integer(), nullable=False),
                    sa.Column('disk_gb', sa.Integer(), nullable=False),
                    sa.Column('amount', sa.Integer(), nullable=False),
                    sa.Column('affinity', sa.Boolean(), nullable=False),
                    sa.Column('flavor_id',
                              sa.String(length=36), nullable=True),
                    sa.Column('aggregate_id', sa.Integer(), nullable=True),
                    sa.Column('server_group_id',
                              sa.String(length=36), nullable=True),
                    sa.ForeignKeyConstraint(['reservation_id'],
                                            ['reservations.id'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('instance_reservations')
    # ### end Alembic commands ###



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_service import threadgroup

from blazar.monitor import base as base_monitor
from blazar.monitor import polling_monitor
from blazar.plugins import base
from blazar import tests


POLLING_INTERVAL = 10
HEALING_INTERVAL = 10


class DummyMonitorPlugin(base.BaseMonitorPlugin):
    def is_notification_enabled(self):
        return True

    def get_notification_event_types(self):
        return []

    def get_notification_topics(self):
        return []

    def notification_callback(self, event_type, message):
        return {}

    def is_polling_enabled(self):
        return False

    def get_polling_interval(self):
        return POLLING_INTERVAL

    def poll(self):
        return {}

    def get_healing_interval(self):
        return HEALING_INTERVAL

    def heal(self):
        return {}


class PollingHandlerTestCase(tests.TestCase):
    def setUp(self):
        super(PollingHandlerTestCase, self).setUp()
        self.monitor_plugins = [DummyMonitorPlugin()]
        self.monitor = polling_monitor.PollingMonitor(self.monitor_plugins)

    def test_start_monitoring(self):
        add_timer = self.patch(threadgroup.ThreadGroup, 'add_timer')
        self.patch(base_monitor.BaseMonitor, 'start_monitoring')

        self.monitor.start_monitoring()
        add_timer.assert_called_once_with(
            POLLING_INTERVAL, self.monitor.call_monitor_plugin, None,
            self.monitor_plugins[0].poll)

    def test_stop_monitoring(self):
        dummy_timer = mock.Mock()
        timer_done = self.patch(threadgroup.ThreadGroup, 'timer_done')
        self.monitor.polling_timers.append(dummy_timer)
        self.patch(base_monitor.BaseMonitor, 'stop_monitoring')

        self.monitor.stop_monitoring()
        timer_done.assert_called_once_with(dummy_timer)



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.tests import api


class TestRoot(api.APITest):
    def setUp(self):
        super(TestRoot, self).setUp()
        self.versions = {
            "versions":
            [{"status": "DEPRECATED",
              "id": "v2.0",
              "links": [{"href": "http://localhost/v2", "rel": "self"}]}]}

    def test_version_discovery_root(self):
        response = self.get_json('/',
                                 expect_errors=True,
                                 path_prefix='')
        self.assertEqual(300, response.status_int)
        self.assertEqual("application/json", response.content_type)
        self.assertEqual(self.versions, response.json)

    def test_version_discovery_versions(self):
        response = self.get_json('/versions',
                                 expect_errors=True,
                                 path_prefix='')
        self.assertEqual(300, response.status_int)
        self.assertEqual("application/json", response.content_type)
        self.assertEqual(self.versions, response.json)

    def test_bad_uri(self):
        response = self.get_json('/bad/path',
                                 expect_errors=True,
                                 path_prefix='')
        self.assertEqual(404, response.status_int)
        self.assertEqual("text/plain", response.content_type)



# Copyright (c) 2020 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.enforcement import filters
from blazar.utils.openstack import base

from oslo_config import cfg
from oslo_log import log as logging

CONF = cfg.CONF

enforcement_opts = [
    cfg.ListOpt('enabled_filters',
                default=[],
                help='List of enabled usage enforcement filters.'),
]

CONF.register_opts(enforcement_opts, group='enforcement')
LOG = logging.getLogger(__name__)


class UsageEnforcement:

    def __init__(self):
        self.load_filters()

    def load_filters(self):
        self.enabled_filters = set()
        for filter_name in CONF.enforcement.enabled_filters:
            _filter = getattr(filters, filter_name)

            if filter_name in filters.all_filters:
                self.enabled_filters.add(_filter(conf=CONF))
            else:
                LOG.error("%s not in filters module.", filter_name)

        self.enabled_filters = list(self.enabled_filters)

    def format_context(self, context, lease_values):
        ctx = context.to_dict()
        region_name = CONF.os_region_name
        auth_url = base.url_for(
            ctx['service_catalog'], CONF.identity_service,
            os_region_name=region_name)

        return dict(user_id=lease_values['user_id'],
                    project_id=lease_values['project_id'],
                    auth_url=auth_url, region_name=region_name)

    def format_lease(self, lease_values, reservations, allocations):
        lease = lease_values.copy()
        lease['reservations'] = []

        for reservation in reservations:
            res = reservation.copy()
            resource_type = res['resource_type']
            res['allocations'] = allocations[resource_type]
            lease['reservations'].append(res)

        return lease

    def check_create(self, context, lease_values, reservations, allocations):
        context = self.format_context(context, lease_values)
        lease = self.format_lease(lease_values, reservations, allocations)

        for _filter in self.enabled_filters:
            _filter.check_create(context, lease)

    def check_update(self, context, current_lease, new_lease,
                     current_allocations, new_allocations,
                     current_reservations, new_reservations):
        context = self.format_context(context, current_lease)
        current_lease = self.format_lease(current_lease, current_reservations,
                                          current_allocations)
        new_lease = self.format_lease(new_lease, new_reservations,
                                      new_allocations)

        for _filter in self.enabled_filters:
            _filter.check_update(context, current_lease, new_lease)

    def on_end(self, context, lease, allocations):
        context = self.format_context(context, lease)
        lease_values = self.format_lease(lease, lease['reservations'],
                                         allocations)

        for _filter in self.enabled_filters:
            _filter.on_end(context, lease_values)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
import oslo_messaging as messaging


CONF = cfg.CONF
TRANSPORT = None


def init():
    global TRANSPORT
    if TRANSPORT is None:
        TRANSPORT = create_transport(get_transport_url())


def get_transport_url(url_str=None):
    return messaging.TransportURL.parse(CONF, url_str)


def get_client(target):
    if TRANSPORT is None:
        raise AssertionError("'TRANSPORT' must not be None")
    return messaging.RPCClient(
        TRANSPORT,
        target,
    )


def get_server(target, endpoints):
    if TRANSPORT is None:
        raise AssertionError("'TRANSPORT' must not be None")
    return messaging.get_rpc_server(
        TRANSPORT,
        target,
        endpoints,
        executor='eventlet',
    )


def create_transport(url):
    return messaging.get_rpc_transport(CONF, url=url)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlalchemy.orm import relationship

from blazar.db.sqlalchemy import model_base as mb

# Helpers


def _generate_unicode_uuid():
    return str(uuidutils.generate_uuid())


def MediumText():
    return sa.Text().with_variant(MEDIUMTEXT(), 'mysql')


def _id_column():
    return sa.Column(sa.String(36),
                     primary_key=True,
                     default=_generate_unicode_uuid)


# Main objects: Lease, Reservation, Event

class Lease(mb.BlazarBase):
    """Contains all info about lease."""

    __tablename__ = 'leases'

    id = _id_column()
    name = sa.Column(sa.String(80), nullable=False)
    user_id = sa.Column(sa.String(255), nullable=True)
    project_id = sa.Column(sa.String(255), nullable=True)
    start_date = sa.Column(sa.DateTime, nullable=False)
    end_date = sa.Column(sa.DateTime, nullable=False)
    trust_id = sa.Column(sa.String(36))
    reservations = relationship('Reservation', cascade="all,delete",
                                backref='lease', lazy='joined')
    events = relationship('Event', cascade="all,delete",
                          backref='lease', lazy='joined')
    status = sa.Column(sa.String(255))
    degraded = sa.Column(sa.Boolean, nullable=False,
                         server_default=sa.false())

    def to_dict(self):
        d = super(Lease, self).to_dict()
        d['reservations'] = [r.to_dict() for r in self.reservations]
        d['events'] = [e.to_dict() for e in self.events]
        return d


class Reservation(mb.BlazarBase):
    """Specifies group of nodes within a cluster."""

    __tablename__ = 'reservations'

    id = _id_column()
    lease_id = sa.Column(sa.String(36),
                         sa.ForeignKey('leases.id'),
                         nullable=False)
    resource_id = sa.Column(sa.String(36))
    resource_type = sa.Column(sa.String(66))
    status = sa.Column(sa.String(13))
    missing_resources = sa.Column(sa.Boolean, nullable=False,
                                  server_default=sa.false())
    resources_changed = sa.Column(sa.Boolean, nullable=False,
                                  server_default=sa.false())
    instance_reservation = relationship('InstanceReservations',
                                        uselist=False,
                                        cascade='all,delete',
                                        backref='reservation',
                                        lazy='joined')
    computehost_reservation = relationship('ComputeHostReservation',
                                           uselist=False,
                                           cascade="all,delete",
                                           backref='reservation',
                                           lazy='joined')
    computehost_allocations = relationship('ComputeHostAllocation',
                                           uselist=True,
                                           cascade="all,delete",
                                           backref='reservation',
                                           lazy='joined')
    floatingip_reservation = relationship('FloatingIPReservation',
                                          uselist=False,
                                          cascade="all,delete",
                                          backref='reservation',
                                          lazy='joined')
    floatingip_allocations = relationship('FloatingIPAllocation',
                                          uselist=True,
                                          cascade="all,delete",
                                          backref='reservation',
                                          lazy='joined')

    def to_dict(self):
        d = super(Reservation, self).to_dict()

        if self.computehost_reservation:

            res = self.computehost_reservation.to_dict()
            d['hypervisor_properties'] = res['hypervisor_properties']
            d['resource_properties'] = res['resource_properties']
            d['before_end'] = res['before_end']

            if res['count_range']:
                try:
                    minMax = res['count_range'].split('-', 1)
                    (d['min'], d['max']) = map(int, minMax)
                except ValueError:
                    e = "Invalid count range: {0}".format(res['count_range'])
                    raise RuntimeError(e)

        if self.instance_reservation:
            ir_keys = ['vcpus', 'memory_mb', 'disk_gb', 'amount', 'affinity',
                       'flavor_id', 'aggregate_id', 'server_group_id',
                       'resource_properties']
            d.update(self.instance_reservation.to_dict(include=ir_keys))

        if self.floatingip_reservation:
            fip_keys = ['network_id', 'amount']
            d.update(self.floatingip_reservation.to_dict(include=fip_keys))

        return d


class Event(mb.BlazarBase):
    """An events occurring with the lease."""

    __tablename__ = 'events'

    id = _id_column()
    lease_id = sa.Column(sa.String(36), sa.ForeignKey('leases.id'))
    event_type = sa.Column(sa.String(66))
    time = sa.Column(sa.DateTime)
    status = sa.Column(sa.String(13))

    def to_dict(self):
        return super(Event, self).to_dict()


class ResourceProperty(mb.BlazarBase):
    """Defines an resource property by resource type."""

    __tablename__ = 'resource_properties'

    id = _id_column()
    resource_type = sa.Column(sa.String(255), nullable=False)
    property_name = sa.Column(sa.String(255), nullable=False)
    private = sa.Column(sa.Boolean, nullable=False,
                        server_default=sa.false())

    __table_args__ = (sa.UniqueConstraint('resource_type', 'property_name'),)

    def to_dict(self):
        return super(ResourceProperty, self).to_dict()


class ComputeHostReservation(mb.BlazarBase):
    """Description

    Specifies resources asked by reservation from
    Compute Host Reservation API.
    """

    __tablename__ = 'computehost_reservations'

    id = _id_column()
    reservation_id = sa.Column(sa.String(36), sa.ForeignKey('reservations.id'))
    aggregate_id = sa.Column(sa.Integer)
    resource_properties = sa.Column(MediumText())
    count_range = sa.Column(sa.String(36))
    hypervisor_properties = sa.Column(MediumText())
    before_end = sa.Column(sa.String(36))

    def to_dict(self):
        return super(ComputeHostReservation, self).to_dict()


class InstanceReservations(mb.BlazarBase):
    """The definition of a flavor of the reservation."""

    __tablename__ = 'instance_reservations'

    id = _id_column()
    reservation_id = sa.Column(sa.String(36), sa.ForeignKey('reservations.id'))
    vcpus = sa.Column(sa.Integer, nullable=False)
    memory_mb = sa.Column(sa.Integer, nullable=False)
    disk_gb = sa.Column(sa.Integer, nullable=False)
    amount = sa.Column(sa.Integer, nullable=False)
    affinity = sa.Column(sa.Boolean, nullable=True)
    resource_properties = sa.Column(MediumText(), nullable=True)
    flavor_id = sa.Column(sa.String(36), nullable=True)
    aggregate_id = sa.Column(sa.Integer, nullable=True)
    server_group_id = sa.Column(sa.String(36), nullable=True)


class ComputeHostAllocation(mb.BlazarBase):
    """Mapping between ComputeHost, ComputeHostReservation and Reservation."""

    __tablename__ = 'computehost_allocations'

    id = _id_column()
    compute_host_id = sa.Column(sa.String(36),
                                sa.ForeignKey('computehosts.id'))
    reservation_id = sa.Column(sa.String(36),
                               sa.ForeignKey('reservations.id'))

    def to_dict(self):
        return super(ComputeHostAllocation, self).to_dict()


class ComputeHost(mb.BlazarBase):
    """Description

    Specifies resources asked by reservation from
    Compute Host Reservation API.
    """

    __tablename__ = 'computehosts'

    id = _id_column()
    vcpus = sa.Column(sa.Integer, nullable=False)
    cpu_info = sa.Column(MediumText(), nullable=False)
    hypervisor_type = sa.Column(MediumText(), nullable=False)
    hypervisor_version = sa.Column(sa.Integer, nullable=False)
    hypervisor_hostname = sa.Column(sa.String(255), nullable=True)
    service_name = sa.Column(sa.String(255), nullable=True)
    memory_mb = sa.Column(sa.Integer, nullable=False)
    local_gb = sa.Column(sa.Integer, nullable=False)
    status = sa.Column(sa.String(13))
    availability_zone = sa.Column(sa.String(255), nullable=False)
    trust_id = sa.Column(sa.String(36), nullable=False)
    reservable = sa.Column(sa.Boolean, nullable=False,
                           server_default=sa.true())
    computehost_extra_capabilities = relationship('ComputeHostExtraCapability',
                                                  cascade="all,delete",
                                                  backref='computehost',
                                                  lazy='joined')
    computehost_resource_inventories = relationship(
        'ComputeHostResourceInventory', cascade="all,delete",
        backref='computehost', lazy='joined')
    computehost_traits = relationship(
        'ComputeHostTrait', cascade="all,delete",
        backref='computehost', lazy='joined')

    def to_dict(self):
        return super(ComputeHost, self).to_dict()


class ComputeHostExtraCapability(mb.BlazarBase):
    """Description

    Allows to define extra capabilities per administrator request for each
    Compute Host added.
    """

    __tablename__ = 'computehost_extra_capabilities'

    id = _id_column()
    computehost_id = sa.Column(sa.String(36), sa.ForeignKey('computehosts.id'))
    property_id = sa.Column(sa.String(36),
                            sa.ForeignKey('resource_properties.id'),
                            nullable=False)
    capability_value = sa.Column(MediumText(), nullable=False)

    def to_dict(self):
        return super(ComputeHostExtraCapability, self).to_dict()


class ComputeHostResourceInventory(mb.BlazarBase):
    __tablename__ = 'computehost_resource_inventory'

    id = _id_column()
    computehost_id = sa.Column(sa.String(36), sa.ForeignKey('computehosts.id'))
    resource_class = sa.Column(sa.String(255), nullable=False)
    total = sa.Column(sa.Integer, nullable=False)
    reserved = sa.Column(sa.Integer, nullable=False)
    min_unit = sa.Column(sa.Integer, nullable=False)
    max_unit = sa.Column(sa.Integer, nullable=False)
    step_size = sa.Column(sa.Integer, nullable=False)
    allocation_ratio = sa.Column(sa.Float, nullable=False)

    def to_dict(self):
        return super(ComputeHostResourceInventory, self).to_dict()


class ComputeHostTrait(mb.BlazarBase):
    __tablename__ = 'computehost_trait'

    id = _id_column()
    computehost_id = sa.Column(sa.String(36), sa.ForeignKey('computehosts.id'))
    trait = sa.Column(sa.String(255), nullable=False)

    def to_dict(self):
        return super(ComputeHostTrait, self).to_dict()


# Floating IP
class FloatingIPReservation(mb.BlazarBase):
    """Description

    Specifies resources asked by reservation from
    Floating IP Reservation API.
    """

    __tablename__ = 'floatingip_reservations'

    id = _id_column()
    reservation_id = sa.Column(sa.String(36), sa.ForeignKey('reservations.id'))
    network_id = sa.Column(sa.String(255), nullable=False)
    amount = sa.Column(sa.Integer, nullable=False)
    required_fips = relationship('RequiredFloatingIP',
                                 cascade='all,delete',
                                 backref='floatingip_reservation',
                                 lazy='joined')

    def to_dict(self, include=None):
        d = super(FloatingIPReservation, self).to_dict(include=include)
        d['required_floatingips'] = [ip['address'] for ip in
                                     self.required_fips]
        return d


class RequiredFloatingIP(mb.BlazarBase):
    """A table for a requested Floating IP.

    Keeps an user requested floating IP address in a floating IP reservation.
    """
    __tablename__ = 'required_floatingips'

    id = _id_column()
    address = sa.Column(sa.String(255), nullable=False)
    floatingip_reservation_id = sa.Column(
        sa.String(36), sa.ForeignKey('floatingip_reservations.id'))


class FloatingIPAllocation(mb.BlazarBase):
    """Mapping between FloatingIP, FloatingIPReservation and Reservation."""

    __tablename__ = 'floatingip_allocations'

    id = _id_column()
    floatingip_id = sa.Column(sa.String(36),
                              sa.ForeignKey('floatingips.id'))
    reservation_id = sa.Column(sa.String(36),
                               sa.ForeignKey('reservations.id'))


class FloatingIP(mb.BlazarBase):
    """A table for Floating IP resource."""

    __tablename__ = 'floatingips'

    id = _id_column()
    floating_network_id = sa.Column(sa.String(255), nullable=False)
    subnet_id = sa.Column(sa.String(255), nullable=False)
    floating_ip_address = sa.Column(sa.String(255), nullable=False)
    reservable = sa.Column(sa.Boolean, nullable=False,
                           server_default=sa.true())

    __table_args__ = (sa.UniqueConstraint('subnet_id', 'floating_ip_address'),)
